# #Typing imports
import typing as typ

#External imports
import itertools,os,math,subprocess,time

#Internal Imports
from catalog.catalog_config import HOSTNAME

"""
General functions for manipulating python objects

General
    identity

Cluster-related
    get_cluster
    get_analysis_folder
    get_sherlock_folder
    launch
    safeMkDir

String Related
    modify_time
    sub_binds
    replacer

List-related
    partition
    chunks
    lol
    flatten
    gcd
    normalize_list

Antiquated
    change_adsorbatedict_to_indice_list
"""
################################################################################

##########
# General
#---------
identity = lambda x : x
negate   = lambda x: -x
true     = lambda x : True
user     = os.environ['USER']

##################
# Cluster related
#-----------------

def running_jobs_sherlock() -> typ.List[str]:
    return subprocess.check_output(['squeue', '-u',user,'-o','%Z']).split()[1:]

def get_cluster() -> str:
    """Get the name of the cluster this command is being executed on"""
    hostname = HOSTNAME.lower() # type: ignore
    if      'sh'     in hostname: return 'sherlock2'
    elif    'gpu-15' in hostname: return 'sherlock2'
    elif    'su'     in hostname: return 'suncat' #important to distinguish suncat2 and 3?
    elif    'kris'   in hostname: return 'kris'
    elif    'local'  in hostname: return 'local'
    else: raise ValueError("was not able to identify the cluster from hostname = %s"%hostname)



def get_analysis_folder(sd : str) -> str:
    """
    Takes either the stordir or the sherlock folder and gives analysis folder
    """
    sd2 = get_sherlock_folder(sd)
    if  'suncat' in sd2: sd2=sd2.replace('suncat_jobs_copy/jobs/','analysis_suncat/')
    elif 'nersc' in sd2: sd2=sd2.replace('nersc_jobs_copy/jobs/','analysis_nersc/')
    else:              sd2=sd2.replace('jobs/','analysis_sherlock/')
    return sd2

def get_sherlock_folder(sd : str) -> str:
    """Takes a storage_directory and gives the path to its copy on sherlock"""
    replace_dict = {'/nfs/slac/g/suncatfs/ksb/share/jobs/' : '/scratch/users/ksb/share/suncat_jobs_copy/jobs/'
                   ,'/global/cscratch1/sd/krisb/share/jobs/':'/scratch/users/ksb/share/nersc_jobs_copy/jobs/'}
    return replacer(str(sd),replace_dict)


def launch() -> None:
    """Tell FireWorks to submit READY jobs to their respective queues"""
    os.system('$SCRATCH/CataLog/fw/launcher.sh')

def safeMkDir(pth     : str
             ,verbose : bool = True
             ) -> None:
    """Make directories even if they already exist"""
    try:            os.mkdir(pth)
    except OSError:
        if verbose: print('directory %s already exists ?!'%pth)

def safeMkDirForce(pth : str)->None:
    """
    Makes a nested directory, even if intermediate links do not yet exist
    """
    components = pth.split('/')
    curr_dir = [components[0]]
    for c in components[1:]:
        curr_dir.append(c)
        safeMkDir('/'+os.path.join(*curr_dir),verbose=False)

def get_hostname_from_fworker(fworker : str) -> str:
    if fworker == 'sherlock2':
        return 'login.sherlock.stanford.edu'
    elif fworker == 'suncat':
        return 'suncatls1.slac.stanford.edu'
    else:
        raise NotImplementedError('please add the fworker to catalog.misc.utilities.get_hostname_from_fworker')

def get_cluster_of_file(file : str) ->str:
    if '/u/if/' in file:
        return 'suncat'
    elif '/scratch/users/':
        return 'sherlock'
    else:
        raise NotImplementedError('Cluster not found')

def read_into_temp(file_to_read : str
                  ) -> str:
    import tempfile
    current_cluster = get_cluster()
    cluster_to_read = get_cluster_of_file(file_to_read)
    new_path        = tempfile.mkstemp()[-1]
    if cluster_to_read == current_cluster:
        return file_to_read
    if cluster_to_read == 'sherlock':
         cluster_host_name = 'login.sherlock.stanford.edu'
    elif cluster_to_read == 'suncat':
         cluster_host_name = 'suncatls1.slac.stanford.edu'
    else:
         raise NotImplementedError('Cluster not valid')
    os.system('scp -q {}@{}:{} {}'.format(user,cluster_host_name,file_to_read,new_path))
    return new_path

#############
# SQL Related
#-----------

################
# String related
#--------------
def modify_time(t          : str
               ,multiplier : float = 2.0
               ) -> str:
    def printTime(floatHours : float) -> str:
        intHours = int(floatHours)
        return "%02d:%02d" % (intHours,(floatHours-intHours)*60)

    """
    Modifies time in either HH:MM::SS or HH:MM format. Min time = 1 hr, max time = 40 hr
    """
    times = [int(x) for x in t.split(':')]
    HHMMSS = len(times) == 3
    tot = times[0]+times[1]/60.0 + (times[2]/3600.0 if HHMMSS else 0)
    return printTime(min(40,math.ceil(multiplier*tot)))+(':00' if HHMMSS else '')

def sub_binds(sql_select : typ.Any) -> None:
    """Prints a sql command in a human-readable way that can be copypasted
    into DB Browswer for SQLite."""

    keywords = ['INNER','FROM','HAVING','WHERE',"GROUP BY",", "]

    (sql_command,binds) = tuple(sql_select)
    sql_command = sql_command.replace('"','`')
    for b in binds: sql_command=sql_command.replace('%s',repr(b),1)

    replace_dict = {x:('\n\t'+x) for x in keywords}

    print('\n\t'+replacer(sql_command,replace_dict)+'\n')

def replacer(s : str
            ,replace_dict : typ.Dict[str,str]
            ) -> str:
    """Executes a series of string replacement operations, specified by a
    dictionary"""
    for k,v in replace_dict.items(): s = s.replace(k,v)
    return s

def merge_dicts(listDicts : typ.List[dict]) -> dict:
    return dict(itertools.chain.from_iterable([x.items() for x in listDicts])) #presumes no overlap in keys

##############
# List related
#-------------
def chunks(l : list, n : int) -> typ.Generator: #Yield successive n-sized chunks from l.
    for i in range(0, len(l), n): yield l[i:i + n]

def flatten(lol : typ.List[list]) -> list:
    return [item for sublist in lol for item in sublist] #flattens a List Of Lists to a list

def gcd(args : typ.List[typ.Any])->typ.Any:
    """
    Greatest common denominator of a list
    """
    if len(args) == 1:  return args[0]
    L = list(args)
    while len(L) > 1:
        a,b = L[len(L) - 2] , L[len(L) - 1]
        L = L[:len(L) - 2]
        while a:   a, b = b%a, a
        L.append(b)
    return abs(b)

def normalize_list(l : list ) -> list:
    """ [a,a,a,a,b,b,c,c] => [a,a,b,c] """
    if len(l)==0: return l
    d   = {x:l.count(x) for x in l}
    div = gcd(list(d.values()))
    norm= [[k]*(v/div) for k,v in d.items()]
    return [item for sublist in norm for item in sublist]



#Typing imports
import typing as typ
if typ.TYPE_CHECKING:
    from catalog.jobs.jobs import Job

#external modules
import json,pickle,copy
import ase # type: ignore
from ase.io import read # type: ignore
from ase.visualize import view

#internal modules
from catalog.jobs.jobs              import Job
from catalog.misc.utilities         import merge_dicts
from catalog.misc.print_parse       import abbreviate_dict
from catalog.jobs.versions          import current_version
import catalog.misc.atoms           as     ma
from catalog.jobs.repeat            import RepeatChecker

checker = RepeatChecker()

class TrajToJob(object):
    """
    Class to easily submit jobs from an existing traj File

    Traj is a initital traj file that will initilize numerous types of job_submiter

    !!!!DEFAULT CALCULATOR VALUES ARE IN easy_makeQEcalc METHOD!!!!

    Currently supported jobkinds:
    relax
    latticeopt
    vcrelax

    Currently supported codes:
    quantumespresso

    Keys required in user_params:
    kpts, xc, fmax, pw

    latticeopt jobs: xtol

    vcrelax jobs: cell_dofree (optional)

    Surface jobs: facet (for 2D materials simply use json.dumps([0,0,0]))
    """
    def __init__(self
                ,traj              : ase.Atoms
                ,jobkind           : str
                ,structure         : str
                ,user_params       : dict           = {}
                ,repeat_checker    : RepeatChecker  = checker
                ) -> None:

                #Check assertions on user inputs
                assert isinstance(traj,ase.Atoms)
                assert jobkind     in ['relax','vcrelax','latticeopt']

                #Set object member data
                self.traj               = traj
                self.structure          = structure
                self.jobkind            = jobkind
                self.repeat_checker     = repeat_checker

                #Initialize the job parameters with the easy_makeQEcalc
                #User must specify the following keys in user_params:
                #kpts, xc, fmax, pw (See easy_makeQEcalc for more detailed explanation of types)
                self.params              = self.easy_makeQEcalc(user_params)
                self.params['inittraj']  = ma.traj_to_json(traj)
                self.params['jobkind']   = jobkind
                self.params['structure'] = structure
                self.job                 = current_version.process_dict(self.params)

    def easy_makeQEcalc(self
                       ,params : dict
                       ) -> dict:
        """returns a dictionary containing all of the ASE calculator parameters
        required to specify a QE calculator object and its associated force minimizer

        USAGE:
        params: A dictionary containing at least the critical calculator parameters
        that must be specified for each calculator (no defaults are specified). If
        non-critical parameters are supplied, the default values will be updated.
        If non-calculator parameters are supplied they will be unchanged

        Required params:
        key                type:                       ex:
        (kpts or kptden)   (int or json.dumps(list)    (4 or json.dumps([12,12,12]))
        xc                 str                          'RPBE'
        fmax               float                        0.05
        pw                 int                          500


        Output:
        merged_params: A dictionary that contains the merged dictionary of the
        originally supplied parameters and the default parameters with their
        updated values
        """

        ###################
        #Default Parameters
        ###################
        #convergence_params
        convergence_params_defaults = {
        'econv':				1e-5
        ,'mixing':				0.1
        ,'nmix':				10
        ,'maxstep':				200
        ,'nbands':				-20
        ,'sigma':				0.1
        }
        #!!!!IMPORTANT CALCULATOR PARAMETERS!!!!
        calculator_params_defaults = {
        'psp':                  'gbrv15pbe'
        ,'dwrat':               10
        ,'dftcode':             'quantumespresso'
        }

        critical_calc_keys = ['kpts','kptden','xc','pw','fmax']
        default_keys = list(convergence_params_defaults.keys())+list(calculator_params_defaults.keys())

        #Check that critical calculator keys were supplied
        assert set(['xc','pw','fmax']) <= set(params.keys()),"Please supply xc, pw, or fmax"
        assert 'kpts' in params.keys() or 'kptden' in params.keys(), "Please supply kpts or kptden"

        #Merge default dictionaries with the supplied critical parameters
        critical_params = {k: params.get(k) for k in critical_calc_keys}
        merged_params = merge_dicts([critical_params,convergence_params_defaults,calculator_params_defaults])
        merged_params.update({k: params.get(k) for k in params.keys() if k not in  critical_calc_keys})
        return merged_params

    def submit(self
              ,check         : bool = True
              ) -> None:
        view(self.traj)
        msg =  abbreviate_dict(self.params)+'\nNew job, sure you want to submit the following? (y/n)'
        if not check or 'y' in input(msg).lower():
            self.job.submit(self.repeat_checker)
        else:
            print('The job was not submitted')



#Typing imports
import typing as typ

#External imports
import qdarkstyle                                 # type: ignore
import sqlparse                                   # type: ignore
import time
from ase.visualize import view                    # type: ignore
from PyQt5 import QtCore, QtGui, QtWidgets, QtSql # type: ignore

#Internal Imports
import catalog.fw.incomplete as inc
from catalog.gui.launch_form import LaunchForm
from catalog.datalog.db_utils import Query
from catalog.misc.sql_utils import *
from catalog.misc.atoms import json_to_traj,new_view

user = os.environ['USER']

class RelaunchGUI(QtWidgets.QWidget):
    def __init__(self
                ,parent          : typ.Any       = None
                ,incomplete_cols : typ.List[str] = ['Status','Cluster','Queue','Time','Nodes','Job Name','created_on','updated_on','Launch Dir']
                ) -> None:
        QtWidgets.QWidget.__init__(self, parent)
        self.setupUi()

        #Build the table
        self.incomplete_cols = incomplete_cols
        self.number_of_completed_jobs_to_show = 0
        self.completed_fwids_to_show          = []

        self.incomplete_rows = self.load_incomplete_jobs()
        if len(self.incomplete_rows) == 0:
            self.set_message('No Incomplete Jobs!\nGet Submitting You Slacker!')

        self._build_table()
        self.regex_combobox.addItems(['fwid']+self.incomplete_cols)
        self._connect_signals()

    def _connect_signals(self) -> None:
        #Connect table clicking
        self.table.clicked.connect(self._table_click)

        #Connect push buttons
        self.push_update.clicked.connect(self._update)
        self.push_quit.clicked.connect(self._quit)
        self.push_view.clicked.connect(self._view)
        self.push_relaunch.clicked.connect(self._relaunch)
        self.push_show_completed.clicked.connect(self._show_completed)
        #Connect RegEx Functions
        self.regex_edit.textChanged.connect(self.on_lineEdit_textChanged)
        self.regex_combobox.currentIndexChanged.connect(self.on_comboBox_currentIndexChanged)
        self.horizontalHeader = self.table.horizontalHeader()
        self.horizontalHeader.sectionClicked.connect(self.on_view_horizontalHeader_sectionClicked)

    def keyPressEvent(self, event : typ.Any)->None:
        k = event.key()
        m = int(event.modifiers())
        if QtGui.QKeySequence(m+k) == QtGui.QKeySequence('Ctrl+W'):
            self._quit(event)
        elif QtGui.QKeySequence(m+k) == QtGui.QKeySequence('Ctrl+C'):
            self._copy()

    def _build_table(self, reload_incomplete_rows : bool = False)->None:
        self.table.setWordWrap(True)
        self.model = QtGui.QStandardItemModel()
        self.model.setHorizontalHeaderLabels(['fwid']+self.incomplete_cols)
        self.update_table_contents(reload_incomplete_rows = reload_incomplete_rows)
        self.proxy = QtCore.QSortFilterProxyModel(self)
        self.proxy.setSourceModel(self.model)
        self.table.setModel(self.proxy)
        self.table.resizeColumnsToContents()
        self.table.setSortingEnabled(True)
        self.table.sortByColumn(0,QtCore.Qt.DescendingOrder)

        #Remove editing capabilities and set only row selection
        self.table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)

    def update_table_contents(self
                             ,reload_incomplete_rows : bool = True
                             ) -> None:
        inc.detect()
        if reload_incomplete_rows:
            self.incomplete_rows = self.load_incomplete_jobs()
        #Name the Rows
        for i_row, row_data in enumerate(self.incomplete_rows):
            #Set last column to be a check box
            self.model.invisibleRootItem().appendRow(
            [QtGui.QStandardItem(str(row_val)) for row_val in row_data])

    def load_incomplete_jobs(self)->list:
        current_incomplete_jobs = inc.table_data(self.incomplete_cols,fw_ids_to_append = self.completed_fwids_to_show)
        if current_incomplete_jobs is None:
            current_incomplete_jobs = []
        return current_incomplete_jobs

    def _table_click(self,event : typ.Any)->None:
        model_index = self.table.model().mapToSource(event)
        self.clicked_model_index = model_index


    def _copy(self)->None:
        sys_clip = QtWidgets.QApplication.clipboard()
        current_cell_selected = self.model.item(self.clicked_model_index.row(),self.clicked_model_index.column()).text()
        sys_clip.setText(current_cell_selected)

    def _update(self,event : typ.Any)->None:
        self.set_message('Updating Table...')
        self.model.clear()
        self._build_table(reload_incomplete_rows = True)
        self.set_message('Table Updated')

    def _view(self,event : typ.Any)->None:
        fwids_to_view     = self.get_selected_fw_ids()
        if len(fwids_to_view)>0:
            fw_dicts          = map(inc.lpad.get_fw_dict_by_id,fwids_to_view)
            atoms_to_view     = [json_to_traj(fwdict['spec']['params']['inittraj']) for fwdict in fw_dicts]
            new_view(atoms_to_view)

    def _quit(self,
              event : typ.Any) -> None:
        """
        Halts the program
        """
        reply = self.check_with_user()
        if reply:
            self.close()
            QtWidgets.QApplication.exit()

    def _relaunch(self,
               event : typ.Any) -> None:
        """
        Halts the program
        """
        fwids_to_relaunch = self.get_selected_fw_ids()
        if self.central_tabWidget.count() == 1:
            if len(fwids_to_relaunch)>0:
                form = self.central_tabWidget.addTab(LaunchForm(fwids_to_relaunch,self),"Launch Form")
                self.central_tabWidget.setCurrentIndex(1)
            else:
                self.set_message('Please select a job to relaunch')
        else:
            self.set_message('Please close the existing Form to launch new jobs')

    def _show_completed(self) -> None:
        dialogbox                                 = QtWidgets.QInputDialog()
        self.number_of_completed_jobs_to_show, ok = dialogbox.getInt(self, 'Complete Job Shower','Select a job type to submit:',0,False)
        if ok:
            self.completed_fwids_to_show          = inc.lpad.get_fw_ids({'state':'COMPLETED'},sort = [('updated_on',-1)],limit =  self.number_of_completed_jobs_to_show)
        self._build_table(reload_incomplete_rows = True)

    def get_selected_fw_ids(self)->typ.List[int]:
        indexes          = self.table.selectedIndexes()
        if len(indexes)>0:
            selected_job_ids = set()                                                # type: ignore
            for index in sorted(indexes):
                model_index = self.table.model().mapToSource(index)
                selected_job_ids.update([int(self.model.item(model_index.row(),0).text())])
            selected_job_ids = list(selected_job_ids)                               # type: ignore
        else:
            selected_job_ids = []
        return selected_job_ids


    @staticmethod
    def check_with_user(message : str = 'Are you sure you want to quit?') -> bool:
        reply = QtWidgets.QMessageBox.question(None, 'Message',
                     message, QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)
        if reply == QtWidgets.QMessageBox.Yes:
            return True
        else:
            return False

    def set_message(self
                   ,text  : str  = ''
                   ,clear : bool = False
                   ) -> None:
        if clear:
            self.message_box.clear()
        self.message_box.appendPlainText(text)

    def setupUi(self) -> None:
        """
        Create the UI for the QueryGUI object
        """
        self.setObjectName("Query")
        self.resize(1900, 732)
        self.setWindowTitle( "Query GUI")

        #Set Fonts for Push Buttons and Titles
        self.header_font = QtGui.QFont()
        self.header_font.setFamily("Helvetica Neue")
        self.header_font.setPointSize(15)

        # self.centralwidget = QtWidgets.QWidget(self)
        # self.centralwidget.setFont(self.header_font)
        # self.centralwidget.setObjectName("centralwidget")
        self.central_layout = QtWidgets.QHBoxLayout(self)
        self.central_layout.setObjectName("central_layout")
        self.central_tabWidget = QtWidgets.QTabWidget()
        self.central_tabWidget.tabBar().hide()
        self.central_layout.addWidget(self.central_tabWidget)

        self.incomplete_jobs_tab        = QtWidgets.QWidget()
        self.incomplete_jobs_tab_layout = QtWidgets.QHBoxLayout(self.incomplete_jobs_tab)

        ##################
        #Create Left Frame and set its layout
        ##################
        self.left_frame = QtWidgets.QFrame()
        self.left_frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        # self.left_frame.setFrameShape(QtWidgets.QFrame.Panel)
        # self.left_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        # self.left_frame.setLineWidth(2)
        # self.left_frame.setMidLineWidth(1)
        self.left_frame_layout = QtWidgets.QVBoxLayout(self.left_frame)

        #Create push button box and fill it with push buttons
        ##################
        self.push_group = QtWidgets.QGroupBox(self.left_frame)
        self.push_group.setFont(self.header_font)

        self.push_group.setObjectName("push_group")
        self.push_group_layout = QtWidgets.QVBoxLayout(self.push_group)
        self.push_group_layout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.push_group_layout.setObjectName("push_group_layout")

        #Create push_update button
        ##################
        self.push_update = QtWidgets.QPushButton(self.push_group)
        self.push_update.setFont(self.header_font)
        self.push_update.setObjectName("push_update")
        self.push_update.setText( "Update Table")
        self.push_group_layout.addWidget(self.push_update)

        #Create push_show_completed button
        ##################
        self.push_show_completed = QtWidgets.QPushButton(self.push_group)
        self.push_show_completed.setFont(self.header_font)
        self.push_show_completed.setObjectName("push_show_completed")
        self.push_show_completed.setText( "Show Completed Jobs")
        self.push_group_layout.addWidget(self.push_show_completed)

        #Create push_lanch button
        ##################
        self.push_relaunch = QtWidgets.QPushButton(self.push_group)
        self.push_relaunch.setFont(self.header_font)
        self.push_relaunch.setObjectName("push_relaunch")
        self.push_relaunch.setText( "Manage Selected Jobs")
        self.push_group_layout.addWidget(self.push_relaunch)


        # #Create push_check_select button
        # ##################
        # self.push_check_select = QtWidgets.QPushButton(self.push_group)
        # self.push_check_select.setFont(self.header_font)
        # self.push_check_select.setObjectName("push_check_select")
        # self.push_check_select.setText( "Check Selection")
        # self.push_group_layout.addWidget(self.push_check_select)
        #
        #Create view button
        ##################
        self.push_view = QtWidgets.QPushButton(self.push_group)
        self.push_view.setFont(self.header_font)
        self.push_view.setObjectName("push_view")
        self.push_view.setText( "View Selection")
        self.push_group_layout.addWidget(self.push_view)

        #Create push_quit button
        ##################
        self.push_quit = QtWidgets.QPushButton(self.push_group)
        self.push_quit.setFont(self.header_font)
        self.push_quit.setObjectName("push_quit")
        self.push_quit.setText( "Quit")
        self.push_group_layout.addWidget(self.push_quit)

        self.left_frame_layout.addWidget(self.push_group)

        #Create RegEx box
        ##################
        self.regex_group = QtWidgets.QGroupBox(self.left_frame)
        self.regex_group.setFont(self.header_font)
        self.regex_group.setTitle("RegEx Filters")
        self.regex_group_layout = QtWidgets.QFormLayout(self.regex_group)
        self.regex_group_layout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.regex_group_layout.setObjectName("regex_group_layout")

        #Create RegEx combobox and editor
        ##################
        self.regex_combobox = QtWidgets.QComboBox()
        self.regex_group_layout.addRow(QtWidgets.QLabel("Column:"), self.regex_combobox)
        self.regex_edit     = QtWidgets.QLineEdit()
        self.regex_group_layout.addRow(QtWidgets.QLabel("Regex String:"),self.regex_edit)
        self.left_frame_layout.addWidget(self.regex_group)

        #Create message box
        ##################
        self.group_message = QtWidgets.QGroupBox(self.left_frame)
        self.group_message.setFont(self.header_font)
        self.group_message.setObjectName("group_message")
        self.group_message.setTitle( "Messages")
        self.group_message_layout = QtWidgets.QVBoxLayout(self.group_message)
        self.group_message_layout.setObjectName("group_message_layout")
        self.message_box = QtWidgets.QPlainTextEdit(self.group_message)
        self.message_box.setReadOnly(True)
        self.message_box.setObjectName("message_box")
        self.group_message.setTitle( "Messages")

        #Add the message and push_button box to the left_frame
        ##################
        self.group_message_layout.addWidget(self.message_box)
        self.left_frame_layout.addWidget(self.group_message)
        self.incomplete_jobs_tab_layout.addWidget(self.left_frame)

        #Create Right Frame and set its layout
        ##################
        self.right_frame = QtWidgets.QFrame()
        self.right_frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        # self.right_frame.setFrameShape(QtWidgets.QFrame.Panel)
        # self.right_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        # self.right_frame.setLineWidth(2)
        # self.right_frame.setMidLineWidth(0)
        # self.right_frame.setObjectName("right_frame")
        self.right_frame_layout = QtWidgets.QVBoxLayout(self.right_frame)
        self.right_frame_layout.setObjectName("right_frame_layout")

        #Create Result Box
        ##################
        self.group_results = QtWidgets.QGroupBox(self.right_frame)
        self.group_results.setFont(self.header_font)
        self.group_results.setAcceptDrops(False)
        self.group_results.setObjectName("group_results")
        self.group_results.setTitle( "Incomplete Jobs")
        self.group_results_layout = QtWidgets.QVBoxLayout(self.group_results)
        # self.group_results_layout.setContentsMargins(5, 5, 5, 5)
        self.group_results_layout.setObjectName("group_results_layout")
        self.group_results.setTitle( "Incomplete Jobs")
        self.table = QtWidgets.QTableView(self.group_results)
        self.table.setFrameShape(QtWidgets.QFrame.NoFrame)
        # self.table.setFrameShadow(QtWidgets.QFrame.Raised)
        self.table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.table.setSortingEnabled(False)
        self.table.setObjectName("table")
        self.group_results_layout.addWidget(self.table)
        self.right_frame_layout.addWidget(self.group_results)
        self.incomplete_jobs_tab_layout.addWidget(self.right_frame)
        # self.right_frame_layout.setStretch(1,1)
        self.incomplete_jobs_tab_layout.setStretch(1,1)

        self.central_tabWidget.addTab(self.incomplete_jobs_tab,"Query GUI")
        # self.setCentralWidget(self.centralwidget)
        QtCore.QMetaObject.connectSlotsByName(self)



    def create_scroll_area(self, n_constraints : int = 3) -> None:
        """
        Creates the scroll area for the SQL form
        Mainly used for the add and remove constraints buttons

        Parameters
        ----------
        n_constraints : int
            number of constraints visible in the ui of the where clause
        """
        #Place the form in the contents of the scroll area
        self.scroll_contents = QtWidgets.QWidget()
        self.scroll_contents.setGeometry(QtCore.QRect(0, 0, 525, 330))
        self.scroll_contents.setObjectName("scroll_contents")
        self.scroll_contents_layout = QtWidgets.QFormLayout(self.scroll_contents)
        self.scroll_contents_layout.setLabelAlignment(QtCore.Qt.AlignCenter)
        self.scroll_contents_layout.setObjectName("scroll_contents_layout")

        #Create Select label
        self.label_select = QtWidgets.QLabel(self.scroll_contents)
        self.label_select.setFont(self.header_font)
        self.label_select.setAlignment(QtCore.Qt.AlignCenter)
        self.label_select.setObjectName("label_select")
        self.label_select.setText( "SELECT")
        self.scroll_contents_layout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_select)

        #Create Select edit box
        self.select_edit = QtWidgets.QLineEdit(self.scroll_contents)
        self.select_edit.setObjectName("select_edit")
        self.select_edit.setText( "*")
        self.scroll_contents_layout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.select_edit)

        #Create Where label
        self.label_where = QtWidgets.QLabel(self.scroll_contents)
        self.label_where.setFont(self.header_font)
        self.label_where.setAlignment(QtCore.Qt.AlignCenter)
        self.label_where.setObjectName("label_where")
        self.label_where.setText( "WHERE")
        self.scroll_contents_layout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_where)

        #Create Constraint fields
        self.n_constraints = 0
        self.constraint_columns    = [] # type: ignore
        self.constraint_conditions = [] # type: ignore
        self.constraint_edits      = [] # type: ignore
        self.layout_constraints    = [] # type: ignore

        default_columns = ['job_id','user','system_type']+['']*(n_constraints-3)
        [self.create_constraint(col) for col in default_columns] # type: ignore

        #Create Group By label
        self.label_group_by = QtWidgets.QLabel(self.scroll_contents)
        self.label_group_by.setFont(self.header_font)
        self.label_group_by.setAlignment(QtCore.Qt.AlignCenter)
        self.label_group_by.setObjectName("label_group_by")
        self.label_group_by.setText( "GROUP BY")
        self.scroll_contents_layout.setWidget(self.scroll_contents_layout.count(), QtWidgets.QFormLayout.LabelRole, self.label_group_by)
        self.group_by_edit = QtWidgets.QLineEdit(self.scroll_contents)
        self.group_by_edit.setObjectName("group_by_edit")
        self.scroll_contents_layout.setWidget(self.scroll_contents_layout.count(), QtWidgets.QFormLayout.FieldRole, self.group_by_edit)

        #Create Limit label
        self.label_limit = QtWidgets.QLabel(self.scroll_contents)
        self.label_limit.setFont(self.header_font)
        self.label_limit.setAlignment(QtCore.Qt.AlignCenter)
        self.label_limit.setObjectName("label_limit")
        self.label_limit.setText( "LIMIT")
        self.scroll_contents_layout.setWidget(self.scroll_contents_layout.count(), QtWidgets.QFormLayout.LabelRole, self.label_limit)
        self.limit_edit = QtWidgets.QLineEdit(self.scroll_contents)
        self.limit_edit.setObjectName("limit_edit")
        self.scroll_contents_layout.setWidget(self.scroll_contents_layout.count(), QtWidgets.QFormLayout.FieldRole, self.limit_edit)
        self.scroll_form.setWidget(self.scroll_contents)

    def create_constraint(self, column_name : str = '') -> None:
        """
        A UI function for making each row for a constraint in the SQL Form.

        Parameters
        ----------
        column_name : str
            name of column to be set in the line edit box of the constraint row
        """
        #Create Column Editor
        self.constraint_columns.append(QtWidgets.QLineEdit(self.scroll_contents))
        self.constraint_columns[self.n_constraints].setAlignment(QtCore.Qt.AlignCenter)
        self.constraint_columns[self.n_constraints].setText(column_name)
        self.constraint_columns[self.n_constraints].setObjectName("constraint_column[self.n_constraints]")
        self.scroll_contents_layout.setWidget(self.n_constraints+3, QtWidgets.QFormLayout.LabelRole, self.constraint_columns[self.n_constraints])

        self.layout_constraints.append(QtWidgets.QHBoxLayout())
        self.layout_constraints[self.n_constraints].setObjectName("layout_constraint_1")

        #Create Condition combo box
        self.constraint_conditions.append(QtWidgets.QComboBox(self.scroll_contents))
        self.constraint_conditions[self.n_constraints].setObjectName("constraint_conditions_{}".format(self.n_constraints))
        combo_box_strings = ["Equal","Not Equal","Like","Not Like","In"]
        [self.constraint_conditions[self.n_constraints].addItem(item) for item in combo_box_strings]
        self.layout_constraints[self.n_constraints].addWidget(self.constraint_conditions[self.n_constraints])

        #Create Constraint Editor
        self.constraint_edits.append(QtWidgets.QLineEdit(self.scroll_contents))
        self.constraint_edits[self.n_constraints].setObjectName("constraint_edits[self.n_constraints]")
        self.layout_constraints[self.n_constraints].addWidget(self.constraint_edits[self.n_constraints])
        self.scroll_contents_layout.setLayout(self.n_constraints+3, QtWidgets.QFormLayout.FieldRole, self.layout_constraints[self.n_constraints])
        self.n_constraints += 1

    @QtCore.pyqtSlot(int)                                            # type: ignore
    def on_view_horizontalHeader_sectionClicked(self, logicalIndex): # type: ignore
        self.logicalIndex   = logicalIndex
        self.menuValues     = QtWidgets.QMenu(self)
        self.signalMapper   = QtCore.QSignalMapper(self)

        self.regex_combobox.blockSignals(True)
        self.regex_combobox.setCurrentIndex(self.logicalIndex)
        self.regex_combobox.blockSignals(False)

        valuesUnique = [    self.model.item(row, self.logicalIndex).text()
                            for row in range(self.model.rowCount())
                            ]
        if len(set(valuesUnique)) == len(self.incomplete_rows):
            return
        actionAll = QtWidgets.QAction("All", self)
        actionAll.triggered.connect(self.on_actionAll_triggered)
        self.menuValues.addAction(actionAll)
        self.menuValues.addSeparator()

        for actionNumber, actionName in enumerate(sorted(list(set(valuesUnique)))):
            action = QtWidgets.QAction(actionName, self)
            self.signalMapper.setMapping(action, actionNumber)
            action.triggered.connect(self.signalMapper.map)
            self.menuValues.addAction(action)

        self.signalMapper.mapped.connect(self.on_signalMapper_mapped)

        headerPos = self.table.mapToGlobal(self.horizontalHeader.pos())

        posY = headerPos.y() + self.horizontalHeader.height()
        posX = headerPos.x() + self.horizontalHeader.sectionPosition(self.logicalIndex)

        self.menuValues.exec_(QtCore.QPoint(posX, posY))

    @QtCore.pyqtSlot()                                         # type: ignore
    def on_actionAll_triggered(self)->None:                    # type: ignore
        filterColumn = self.logicalIndex
        filterString = QtCore.QRegExp( "",
                                        QtCore.Qt.CaseInsensitive,
                                        QtCore.QRegExp.RegExp
                                        )
        self.proxy.setFilterRegExp(filterString)
        self.proxy.setFilterKeyColumn(filterColumn)

    @QtCore.pyqtSlot(int)                                      # type: ignore
    def on_signalMapper_mapped(self, i : typ.Any)->None:       # type: ignore
        stringAction = self.signalMapper.mapping(i).text()
        filterColumn = self.logicalIndex
        filterString = QtCore.QRegExp( stringAction,
                                        QtCore.Qt.CaseSensitive,
                                        QtCore.QRegExp.FixedString
                                        )
        self.proxy.setFilterRegExp(filterString)
        self.proxy.setFilterKeyColumn(filterColumn)

    @QtCore.pyqtSlot(str)                                      # type: ignore
    def on_lineEdit_textChanged(self, text : typ.Any) -> None: # type: ignore
        search = QtCore.QRegExp( text,
                                    QtCore.Qt.CaseInsensitive,
                                    QtCore.QRegExp.RegExp
                                    )

        self.proxy.setFilterRegExp(search)

    @QtCore.pyqtSlot(int)
    def on_comboBox_currentIndexChanged(self, index):          # type: ignore
        self.proxy.setFilterKeyColumn(index)




if __name__ == "__main__":
    import sys,os
    os.environ['QT_API']='pyqt5'
    app = QtWidgets.QApplication(sys.argv)
    app.setStyleSheet(qdarkstyle.load_stylesheet_from_environment(is_pyqtgraph=False))

    gui = RelaunchGUI()
    gui.showMaximized()
    sys.exit(app.exec_())



# Typing Modules
import typing as typ

# #External Modules
import os,json
from pprint import pformat
from time import sleep

#--sql imports
import sql                                                      # type: ignore
from sql import As,Table,Join,Flavor,Literal                    # type: ignore
from sql.operators import In,And,Or,Not,Concat,Mod,Like,NotLike # type: ignore
from sql.functions import Function,Substring,Round,Random       # type: ignore

from MySQLdb import connect,Connection,OperationalError         # type: ignore
from MySQLdb.cursors import Cursor,DictCursor                   # type: ignore

################################################################################

# ######################
# # Connection for the MySQL database
# # --------------------

localuser = os.environ["USER"]

class ConnectInfo(object):
    """MySQL connection info """
    def __init__(self
                ,host   : str = '127.0.0.1'
                ,port   : int = 3306
                ,user   : str = localuser
                ,passwd : str = localuser
                ,db     : str = 'suncat'
                ) -> None:
        self.host   = host
        self.port   = port
        self.user   = user
        self.passwd = passwd
        self.db     = db

    def __str__(self)->str:
        return pformat(self.__dict__)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__

    def mk_conn(self
               ,mk_dict : bool = False
               ,attempt : int  = 10
               ) -> Connection:
        try:
            con = connect(host   = self.host
                          ,port   = self.port
                          ,user   = self.user
                          ,passwd = self.passwd
                          ,db     = self.db
                          ,cursorclass = DictCursor if mk_dict else Cursor
                          ,connect_timeout=28800)
                          #,wait_timeout = 28800
                          #,interactive_timeout = 28800)

        except OperationalError as e:
            if attempt > 0:
                sleep(5)
                con = self.mk_conn(mk_dict,attempt-1)
            else:
                raise OperationalError(e)
        con.query('SET SESSION wait_timeout=28800')
        return con

#####################
# Auxillary Functions
#--------------------
def joiner_simple(first_table : sql.Table
                 ,dictionary  : dict
                 ) -> sql.Join:
    """
    Concisely express a join
    This function is needed because Mike's joiner breaks when your join condition
    is itself a list (i.e. joining on 2+ conditions)

    New format for specifying join type: dictionary value is a TUPLE (c,t)
    """
    output = first_table
    for t,c in dictionary.items():
        if isinstance(c,tuple):            c,typ = c
        else: typ = 'INNER'
        output=Join(output,t,condition=c,type_ = typ)
    return output

def joiner(first_table      : sql.Table
          ,subsequent_dict  : dict
          ,join_type        : str           = 'INNER'
          ) -> sql.Join:
    """
    Concisely express a join

    no longer supports 'subsequent_dict' being a list
    """
    output = first_table

    for t,c in subsequent_dict.items():
        if isinstance(c,list) and not isinstance(c,sql.operators.And) and len(c)>1:
            output=Join(output,t,condition=c[0], type_ = c[1])
        else:
            output=Join(output,t,condition=c, type_ = join_type)
    return output

def join_joins(first_join   : typ.Any
              ,second_join  : typ.Any
              ,condition    : typ.Any
              ,join_type    : str       =  'INNER'
              ) -> typ.Any:
    """
    Typing was probably done completely wrong
    """
    from catalog.misc.utilities import merge_dicts
    first_main_table, first_dictionary = unpack_joins(first_join)
    second_main_table, second_dictionary  = unpack_joins(second_join)
    new_condition = {second_main_table : [condition, join_type]}
    return joiner(first_main_table, [new_condition,first_dictionary, second_dictionary])  # type: ignore

def unpack_joins(joined_tables : typ.Any
                 ) -> typ.Any:
    """
    Turns a join statement into the input used to create it
    """
    from catalog.misc.utilities import merge_dicts
    left        = joined_tables.left
    right       = joined_tables.right
    condition   = joined_tables.condition
    join_type   = joined_tables.type_
    if isinstance(left,Table) and isinstance(right,Table):
        return [left, {right:[condition, join_type]}]
    elif isinstance(left,Table):
        output = unpack_joins(right)
        output[1] = merge_dicts([output[1], {left: [condition, join_type]}])
        return output
    else:
        output = unpack_joins(left)
        output[1] = merge_dicts([output[1], {right: [condition, join_type]}])
        return output

def table_names(joined_tables : typ.Any
                )->typ.List[str]:
    """
    Turns a join statement into the input used to create it
    """
    if isinstance(joined_tables,Table):
        return [joined_tables._name]

    left        = joined_tables.left
    right       = joined_tables.right
    if isinstance(left,Table) and isinstance(right,Table):
        return [left._name, right._name]
    elif isinstance(left,Table):
        output = table_names(right)
        output += [left._name]
        return output
    else:
        output = table_names(left)
        output += [right._name]
        return output

def concat(*args : typ.Any) -> sql.operators.Concat:
    """
    Concat a list of objects
    """
    output = Concat(*args[-2:])
    for a in reversed(args[:-2]): output=Concat(a,output)
    return output

def AND(*args : typ.Any) -> sql.operators.And:
    args = list(filter(None,args)) # type: ignore
    if len(args)==0: return None
    return And(args)

def OR(*args : typ.Any) -> sql.operators.Or:
    # args = [a for a in args if a not in [None,[]]]  # type: ignore
    # if len(args)==0: return None
    return Or(args)

#############
# Function Definitions
#------------
class UDF(Function):         __slots__ = () ; _function = 'user_defined_function'
class ABS(Function):         __slots__ = () ; _function = 'abs'
class SQRT(Function):        __slots__ = (); _function = 'sqrt'
class SUBSTR(Function):      __slots__ = () ; _function = 'substr'
class IFNULL(Function):      __slots__ = () ; _function = 'ifnull'
class GROUPCONCAT(Function): __slots__ = () ; _function = 'group_concat'
class CON(Function):      __slots__ = () ; _function = 'concat'
class DISTINCT(Function):    __slots__ = () ; _function = 'Distinct'
class IIF(Function):         __slots__ = () ; _function = 'IIF'
class FLOAT(Function):         __slots__ = () ; _function = '1.0*'

#########
# Tables
#-------
tbl_adsorbate               = Table('adsorbate')
tbl_adsorbate_composition   = Table('adsorbate_composition')
tbl_ads_triple              = Table('ads_triple')
tbl_alljob                  = Table('alljob')
tbl_atom                    = Table('atom')
tbl_bond                    = Table('bond')
tbl_bulk                    = Table('bulk')
tbl_bulkmod_job             = Table('bulkmod_job')
tbl_bulkspecies             = Table('bulkspecies')
tbl_calc                    = Table('calc')
tbl_calc_other              = Table('calc_other')
tbl_cell                    = Table('cell')
tbl_chargemol_atom          = Table('chargemol_atom')
tbl_chargemol_job           = Table('chargemol_job')
tbl_chargemol_map           = Table('chargemol_map')
tbl_dos_job                 = Table('dos_job')
tbl_element                 = Table('element')
tbl_graph                   = Table('graph')
tbl_graph_params            = Table('graph_params')
tbl_hyperparameter          = Table('hyperparameter')
tbl_yperparameter_type      = Table('yperparameter_type')
tbl_hyperparameter_value    = Table('hyperparameter_value')
tbl_keld_solids             = Table('keld_solids')
tbl_META_dependencies       = Table('META_dependencies')
tbl_ETA_transform           = Table('ETA_transform')
tbl_molecule                = Table('molecule')
tbl_molspecies              = Table('molspecies')
tbl_neb_job                 = Table('neb_job')
tbl_pure_struct             = Table('pure_struct')
tbl_ref_scheme              = Table('ref_scheme')
tbl_ref_scheme_elem         = Table('ref_scheme_elem')
tbl_ref_scheme_job          = Table('ref_scheme_job')
tbl_ref_scheme_refstruct    = Table('ref_scheme_refstruct')
tbl_ref_scheme_reqs         = Table('ref_scheme_reqs')
tbl_ref_scheme_stoich       = Table('ref_scheme_stoich')
tbl_relax_job               = Table('relax_job')
tbl_roots                   = Table('roots')
tbl_rxn                     = Table('rxn')
tbl_rxn_dataset             = Table('rxn_dataset')
tbl_rxn_dataset_element     = Table('rxn_dataset_element')
tbl_similar_struct          = Table('similar_struct')
tbl_species                 = Table('species')
tbl_species_composition     = Table('species_composition')
tbl_stoich                  = Table('stoich')
tbl_struct                  = Table('struct')
tbl_struct_adsorbate        = Table('struct_adsorbate')
tbl_struct_composition      = Table('struct_composition')
tbl_struct_dataset          = Table('struct_dataset')
tbl_truct_dataset_element   = Table('truct_dataset_element')
tbl_surface                 = Table('surface')
tbl_surfspecies             = Table('surfspecies')
tbl_traj                    = Table('traj')
tbl_trajatom                = Table('trajatom')
tbl_vib_job                 = Table('vib_job')
tbl_finaltraj               = Table('finaltraj')
tbl_job                     = Table('job')

# ######################
# # Table Joins
# # --------------------
tbl_relaxfinal = joiner(tbl_relax_job,{
                         tbl_job                 : tbl_job.job_id                     == tbl_relax_job.job_id
                        ,tbl_finaltraj           : tbl_finaltraj.job_id               == tbl_relax_job.job_id
                        ,tbl_struct              : tbl_struct.struct_id               == tbl_finaltraj.struct_id
                        ,tbl_calc                : tbl_relax_job.calc_id              == tbl_calc.calc_id
                        ,tbl_calc_other          : tbl_relax_job.calc_other_id        == tbl_calc_other.calc_other_id
                        })

tbl_relaxfinal_surface   = joiner(tbl_relaxfinal,{
                         tbl_surface : tbl_struct.struct_id == tbl_surface.struct_id
                         })

tbl_relaxfinal_bulk      = joiner(tbl_relaxfinal,{
                         tbl_bulk : tbl_struct.struct_id == tbl_bulk.struct_id
                         })

tbl_relaxfinal_molecule  = joiner(tbl_relaxfinal,{
                         tbl_molecule : tbl_struct.struct_id == tbl_molecule.struct_id
                         })

tbl_relaxfinal_all       = joiner(tbl_relaxfinal,{
                          tbl_molecule : [tbl_struct.struct_id == tbl_molecule.struct_id, 'LEFT']
                         ,tbl_bulk     : [tbl_struct.struct_id == tbl_bulk.struct_id, 'LEFT']
                         ,tbl_surface  : [tbl_struct.struct_id == tbl_surface.struct_id, 'LEFT']
                         })

# ######################
# # Adsorption Query
# # --------------------
T       = Table('ads_triple')
BF      = Table('finaltraj')
CF      = Table('finaltraj')
BR      = Table('relax_job')
CR      = Table('relax_job')
BJ      = Table('job')
CJ      = Table('job')
BS      = Table('struct')
CS      = Table('struct')
BC      = Table('calc_other')
CC      = Table('calc_other')
A       = Table('ads_eng')
AF      = Table('finaltraj')
AS      = Table('struct')
AJ      = Table('job')
AC      = Table('calc_other')

tbl_eads       = joiner(T, {
                        BF : BF.job_id           == T.bare_id
                       ,CF : CF.job_id           == T.complex_id
                       ,BR : BR.job_id           == T.bare_id
                       ,CR : CR.job_id           == T.complex_id
                       ,BJ : BJ.job_id           == T.bare_id
                       ,CJ : CJ.job_id           == T.complex_id
                       ,BS : BS.struct_id        == BF.struct_id
                       ,CS : CS.struct_id        == CF.struct_id
                       ,BC : BC.calc_other_id    == BR.calc_other_id
                       ,CC : CC.calc_other_id    == CR.calc_other_id
                       ,A  : And([A.adsorbate_id == T.adsorbate_id, A.calc_id == T.calc_id])
                       ,AJ : AJ.job_id           == A.job_id
                       ,AC : AC.calc_other_id    == A.calc_other_id
                       ,AF : AF.job_id           == AJ.job_id
                       ,AS : AS.struct_id        == AF.struct_id
                       })

eads_columns_inner  = [T.bare_id
                      ,T.complex_id
                      ,T.delta_e_surf
                      ,As(CJ.job_name,'complex_name')
                      ,As(BJ.job_name,'bare_name')
                      ,A.name
                      ,A.ref_scheme_id
                      ,A.energy_norm
                      ,A.coef
                      ,A.number
                      ,A.element_id
                      ,A.component
                      ,As(GROUPCONCAT(A.job_id),'jobs')]

eads_group_by_inner = [A.ref_scheme_id
                      ,A.calc_id
                      ,A.adsorbate_id
                      ,A.element_id
                      ,A.component_id
                      ,T.complex_id
                      ,T.bare_id]

###############
# Col shortcuts
# -------------
#--job table
JOB                         = tbl_job.job_id
USER                        = tbl_job.user
JOBCODE                     = tbl_job.code
DELETED                     = tbl_job.deleted
JOBNAME                     = tbl_job.job_name
JOBTYPE                     = tbl_job.job_type
LOGFILE                     = tbl_job.logfile
STORDIR                     = tbl_job.stordir
TIMESTAMP                   = tbl_job.timestamp
WORKDIR                     = tbl_job.working_directory
ADSORBATES                  = tbl_job.ads_catalog
STRUCTURE                   = tbl_job.structure_catalog

#--relax_job table
RELAX_JOB                   = tbl_relax_job.job_id
RELAX_CALC                  = tbl_relax_job.calc_id
RELAX_CALC_OTHER            = tbl_relax_job.calc_other_id
REFERENCE                   = tbl_relax_job.calc_other_id

#--traj table
TRAJ                        = tbl_traj.traj_id
TRAJ_JOB                    = tbl_traj.job_id
TRAJ_STRUCT                 = tbl_traj.struct_id
FINAL                       = tbl_traj.final
RAWENG                      = tbl_finaltraj.energy
KPTDEN_X                    = tbl_traj.kptden_x
KPTDEN_Y                    = tbl_traj.kptden_y
KPTDEN_Z                    = tbl_traj.kptden_z

#--struct table
STRUCT                      = tbl_struct.id
STRUCT_CELL                 = tbl_struct.cell_id
RAW                         = tbl_struct.raw
SYSTYPE                     = tbl_struct.system_type
NATOMS                      = tbl_struct.n_atoms
NELEMS                      = tbl_struct.n_elems
# struct columns to be aliased
#                           = tbl_struct.composition
#                           = tbl_struct.composition_norm
#                           = tbl_struct.metal_comp
#                           = tbl_struct.str_symbols
#                           = tbl_struct.str_constraints
#                           = tbl_struct.symmetry
#                           = tbl_struct.pure_struct_id
#                           = tbl_struct.species_id
#                           = tbl_struct.spacegroup
#                           = tbl_struct.geo_graph
#                           = tbl_struct.elemental
#                           = tbl_struct.pure_struct_id_catalog
#                           = tbl_struct.rawhash

#--surface table
SURFACE                     = tbl_surface.struct_id
VACUUM                      = tbl_surface.vacuum
SURFACE_PURE_STRUCT         = tbl_surface.pure_struct_id
H                           = tbl_surface.facet_h
K                           = tbl_surface.facet_k
L                           = tbl_surface.facet_l
SURFACE_PURE_STRUCT_CATALOG = tbl_surface.pure_struct_id_catalog
H_CATALOG                   = tbl_surface.facet_h_catalog
K_CATALOG                   = tbl_surface.facet_k_catalog
L_CATALOG                   = tbl_surface.facet_l_catalog
FACET                       = CON('[',H,',',K,',',L,']')
FACET_CATALOG               = CON('[',H_CATALOG,',',K_CATALOG,',',L_CATALOG,']')


#--pure_struct table
PURE_STRUCT                 = tbl_pure_struct.pure_struct_id
NAME                        = tbl_pure_struct.name
SPACEGROUP                  = tbl_pure_struct.spacegroup
FREE                        = tbl_pure_struct.free
NICKNAME                    = tbl_pure_struct.nickname



#--calc table
CALC                = tbl_calc.id
DFTCODE             = tbl_calc.dftcode
XC                  = tbl_calc.xc
PW                  = tbl_calc.pw
PSP                 = tbl_calc.psp

#--calc_other table
CALC_OTHER          = tbl_calc_other.calc_other_id
KX                  = tbl_calc_other.kx
KY                  = tbl_calc_other.ky
KZ                  = tbl_calc_other.kz
FMAX                = tbl_calc_other.fmax
ECONV               = tbl_calc_other.econv
XTOL                = tbl_calc_other.xtol
STRAIN              = tbl_calc_other.strain
DW                  = tbl_calc_other.dw
SIGMA               = tbl_calc_other.sigma
NBANDS              = tbl_calc_other.nbands
MIXING              = tbl_calc_other.mixing
NMIX                = tbl_calc_other.nmix
GGA                 = tbl_calc_other.gga
LUSE_VDW            = tbl_calc_other.luse_vdw
ZAB_VDW             = tbl_calc_other.zab_vdw
NELMDL              = tbl_calc_other.nelmdl
GAMMA               = tbl_calc_other.gamma
DIPOL               = tbl_calc_other.dipol
ALGO                = tbl_calc_other.algo
IBRION              = tbl_calc_other.ibrion
PREC                = tbl_calc_other.prec
IONIC_STEPS         = tbl_calc_other.ionic_steps
LREAL               = tbl_calc_other.lreal
LVHAR               = tbl_calc_other.lvhar
DIAG                = tbl_calc_other.diag
SPINPOL             = tbl_calc_other.spinpol
DIPOLE              = tbl_calc_other.dipole
MAXSTEP             = tbl_calc_other.maxstep
DELTA               = tbl_calc_other.delta
MIXINGTYPE          = tbl_calc_other.mixingtype
BONDED_INDS         = tbl_calc_other.bonded_inds
ENERGY_CUTOFF       = tbl_calc_other.energy_cut_off
STEP_SIZE           = tbl_calc_other.step_size
SPRINGS             = tbl_calc_other.springs



ADSID               = tbl_atom.adsorbate_id
ADSID_CATALOG       = tbl_atom.adsorbate_id_catalog
ATOM                = tbl_atom.atom_id
CONSTRAINED         = tbl_atom.constrained
ATOM_TO_ELEMENT     = tbl_atom.element_id
LAYER               = tbl_atom.layer
MAGMOM              = tbl_atom.magmom
ATOM_TO_PURE_STRUCT = tbl_atom.pure_struct_id
ATOM_STRUCT_ID      = tbl_atom.struct_id
X                   = tbl_atom.x
Y                   = tbl_atom.y
Z                   = tbl_atom.z

# ######################
# # Constraint shortcuts
# # --------------------
JOB_          = lambda x: JOB       == x
JOB_IN_       = lambda x: In(JOB,x)
JOBNAME_      = lambda x: JOBNAME   == x
JOBNAME_LIKE_ = lambda x: Like(JOBNAME,x)

ATOM_         = lambda x: ATOM      == x
STRUCT_       = lambda x: STRUCT    == x
CALC_         = lambda x: CALC      == x
STRUCTURE_    = lambda x: STRUCTURE == x
STORDIR_      = lambda x: STORDIR == x                      # type: ignore
JOBNAME_      = lambda x: JOBNAME == x                      # type: ignore
#
USER_    = lambda x: USER == x                              # type: ignore
KSB_     = USER_('ksb')                                     # type: ignore
MSTATT_  = USER_('mstatt')                                  # type: ignore
AAYUSH_  = USER_('aayush')                                  # type: ignore

def JOBTYPE_(x): return JOBTYPE == x                        # type: ignore
LATTICEOPT_ = JOBTYPE_('latticeopt')                        # type: ignore
BULKMOD_    = JOBTYPE_('bulkmod')                           # type: ignore
VIB_        = JOBTYPE_('vib')                               # type: ignore
VCRELAX_    = JOBTYPE_('vcrelax')                           # type: ignore
RELAX_      = JOBTYPE_('relax')                             # type: ignore
DOS_        = JOBTYPE_('dos')                               # type: ignore
NEB_        = JOBTYPE_('neb')                               # type: ignore
XCCONTRIBS_ = JOBTYPE_('xc')                                # type: ignore

RELAXORLAT_  = OR(LATTICEOPT_,VCRELAX_,RELAX_)

def NATOMS_(x): return NATOMS == x                          # type: ignore
PW_ = lambda x: PW==x                                       # type : ignore
def XC_(x : str) -> bool: return XC == x                    # type : ignore
PBE_   = XC_('PBE')                                         # type : ignore
RPBE_  = XC_('RPBE')                                        # type : ignore
BEEF_  = XC_('BEEF')                                        # type : ignore
MBEEF_ = XC_('mBEEF')                                       # type : ignore

def DFTCODE_(x): return DFTCODE==x                          # type: ignore
GPAW_ = DFTCODE_('gpaw')                                    # type: ignore
QE_ = DFTCODE_('quantumespresso')                           # type: ignore
VASP_ = DFTCODE_('vasp')                                    # type: ignore

def NICKNAME_(x): return NICKNAME == x                    # type: ignore
HCP_     = NICKNAME_('hexagonal')                          # type: ignore
FCC_     = NICKNAME_('fcc')                                # type: ignore
BCC_     = NICKNAME_('bcc')                                # type: ignore
DIAMOND_ = NICKNAME_('diamond')                            # type: ignore
#
def SYSTYPE_(x): return SYSTYPE == x                        # type: ignore
SURFACE_     = SYSTYPE_('surface')                          # type: ignore
BULK_        = SYSTYPE_('bulk')                             # type: ignore
MOLECULE_    = SYSTYPE_('molecule')                         # type: ignore

def FACET_(x): return FACET == json.dumps(x)                 # type: ignore
F111_ = FACET_([1,1,1])                                      # type: ignore
F110_ = FACET_([1,1,0])                                      # type: ignore
F001_ = FACET_([0,0,1])                                      # type: ignore


# ######################
# # Archived Scripts
# # --------------------

# #################
# # Join conditions
# #---------------
# atom__job           = atom.atoms_id == job.finalatoms
# atom__atoms         = atom.atoms_id == atoms.id
# atom__refstoich     = atom.number   == refstoich.reference_element
# atom__element       = atom.number   == element.id
#
# calc__job           = calc.id       == job.calc_id
# calc__refeng        = calc.id       == refeng.calc_id
#
# atoms__job          = atoms.id      == job.finalatoms
# atoms__cell         = cell.id       == atoms.cell_id
# atom__refeng        = atom.number   == refeng.element_id
#
# atoms__composition  = atoms.id               == composition.atoms_id
# composition__job    = composition.atoms_id   == job.finalatoms
# composition__refeng = composition.element_id == refeng.element_id
#
# init_or_finalatoms  = OR(job.finalatoms  == atoms.id, job.initatoms   == atoms.id)
# init_or_finalatoms2 = OR(job.finalatoms  == atom.atoms_id
#                         ,job.initatoms   == atom.atoms_id)
#
# atom__composition   = AND(atom.atoms_id  == composition.atoms_id
#                          ,atom.number    == composition.element_id)
#
# Columns to be aliased
# = tbl_atom.struct_adsorbate_id
# = tbl_atom.struct_adsorbate_id_catalog
# = tbl_atom.tag
# = tbl_atom.v10
# = tbl_atom.v3
# = tbl_atom.v4
# = tbl_atom.v5
# = tbl_atom.v6
# = tbl_atom.v7
# = tbl_atom.v8
# = tbl_atom.v9

# CALC_ID     = tbl_job.calc_id
# INITATOMS   = tbl_job.initatoms
# FINALATOMS  = tbl_job.finalatoms
####

# KPTDEN_X    = tbl_job.kptden_x
# KPTDEN_Y    = tbl_job.kptden_y
# KPTDEN_Z    = tbl_job.kptden_z
# PARENT      = tbl_job.parent
# BULKMODULUS = tbl_job.bulk_modulus
# ADSORBATES  = tbl_job.adsorbates
# STRJOB_EXACT = tbl_job.strjob_exact
# STRJOB_GRAPH = tbl_job.strjob_graph
#
# CHARGEMOL   = tbl_job.chargemol
# GRAPH       = tbl_job.graph
# FWID        = tbl_job.fwid
# BLANK       = tbl_job.blank
# RAWENG      = tbl_job.raw_energy
# SURFENG     = tbl_job.surface_energy
# RAWG        = tbl_job.raw_g
# REFJOB      = tbl_job.refjob
# EFORM       = tbl_job.eform
# GFORM       = tbl_job.gform
# EADS        = tbl_job.e_ads
# GADS        = tbl_job.g_ads
# VIBFREQS    = tbl_job.vib_freqs
####################
#
# ATOM        = atom.id
# IND         = atom.ind
# NUMBER      = atom.number
# X           = atom.x
# Y           = atom.y
# Z           = atom.z
# CONSTRAINED = atom.constrained
# MAGMOM      = atom.magmom
# TAG         = atom.tag
# ATOMSID     = atom.atoms_id
# ADSORBATE   = atom.adsorbate
# ####
# COORDNUM    = atom.coordination_number
# Q4          = atom.q4
# Q6          = atom.q6



#
# ####################




# ###
# NATOMS      = atoms.natoms
# NATOMS_CONSTRAINED = atoms.natoms_constrained
# SYSTYPE     = atoms.system_type
# POINTGROUP  = atoms.pointgroup
# SPACEGROUP  = atoms.spacegroup
# STRUCTURE   = atoms.structure
# FACET       = atoms.facet
# SURFACEAREA = atoms.surface_area
# VOLUME      = atoms.volume
# HAS_ADS     = atoms.has_ads
# SYMSLAB     = atoms.sym_slab
# ADS_NELEMS  = atoms.ads_nelems
# NELEMS      = atoms.nelems


# ####################
# CELL  = cell.id
# AX    = cell.ax
# AY    = cell.ay
# AZ    = cell.az
# BX    = cell.bx
# BY    = cell.by
# BZ    = cell.bz
# CX    = cell.cx
# CY    = cell.cy
# CZ    = cell.cz
# ####################
# #####
# ELEMENT       = element.id
# REFSPACEGROUP = element.reference_spacegroup
# REFENG        = refeng.id
# ALL_EREF      = refeng.all_eref
# ALL_GREF      = refeng.all_gref
# EREF          = refeng.eref
# GREF          = refeng.gref
# EATOM         = refeng.eatom
# GATOM         = refeng.gatom
# #############
# REFSTOICH     = refstoich.id
# REF_ELM       = refstoich.reference_element
# COMPONENT_ELM = refstoich.component_element
# COMPWEIGHT    = refstoich.component_weight
# TOT_COMPS     = refstoich.total_components
# ########
# COMPOSITION   = composition.id
# HAS           = composition.has
# CONST_HAS     = composition.const_has
# COUNT         = composition.count
# CONST_COUNT   = composition.const_count
# ADS_COUNT     = composition.ads_count
# FRAC          = composition.frac
# CONST_FRAC    = composition.const_frac
#
#


# NUMBER_    = lambda x: NUMBER    == x
#
# ADSORBATE_ = lambda x: ADSORBATE == x
# ADSORBATES_ = lambda x: ADSORBATES == x
# BARE_ = ADSORBATES_('[]') # type: ignore
#
# FWID_    = lambda x: FWID == x
# FWIDS_   = lambda xs: In(FWID,tuple(xs))
#
# STRJOB_EXACT_ = lambda x: STRJOB_EXACT == x
# STRJOB_GRAPH_ = lambda x: STRJOB_GRAPH == x
# GRAPH_        = lambda x: GRAPH == x

# def NATOMS_CONSTRAINED_(x): return NATOMS_CONSTRAINED == x  # type: ignore
#

#
# def PSP_(x): return PSP == x                                # type: ignore
# SG15_   = PSP_('sg15')                                      # type: ignore
# GBRV_   = PSP_('gbrv15pbe')                                 # type: ignore
# PAW_    = PSP_('paw')                                       # type: ignore
# OLDPAW_ = PSP_('oldpaw')                                    # type: ignore
#
# KPTS_    = lambda x: KPTS==x                                # type: ignore
# KPTLOW_  = KPTDEN_X < 4
# KPTHIGH_ = Not(KPTLOW_)
#
# ECONV_ = lambda x: ECONV == x
# MIXING_ = lambda x: MIXING == x
# NMIX_   = lambda x: NMIX == x
# FMAX_   = lambda x: FMAX == x
# SPINPOL_ = lambda x: SPINPOL==x
# PARENT_  = lambda x: PARENT == x

#
# cp = {FCC_:F111_, BCC_:F110_, HCP_:F001_}
#
# CLOSEPACKED_ = Or([And([st,f]) for st,f in cp.items()])
#
#
# relevent_atoms = [1,3,4,6,7,8,9,11,12,13,14,16,17] \
#                     + list(range(19,36))+list(range(37,52))+[55,56]+list(range(72,80))
#
# H2_ = REFJOB == 1
# LI_ = REFJOB == 3
# BE_ = REFJOB == 4
# CO_ = REFJOB == 6
# N2_ = REFJOB == 7
# H2O_= REFJOB == 8
# F2_ = REFJOB == 9
# NA_ = REFJOB == 11
# MG_ = REFJOB == 12
# AL_ = REFJOB == 13
# SI_ = REFJOB == 14
# CL2_= REFJOB == 17
# K_  = REFJOB == 19
# CA_ = REFJOB == 20
# SC_ = REFJOB == 21
# TI_ = REFJOB == 22
# V_  = REFJOB == 23
# CR_ = REFJOB == 24
# MN_ = REFJOB == 25
# FE_ = REFJOB == 26
# CO_ = REFJOB == 27
# NI_ = REFJOB == 28
# CU_ = REFJOB == 29
# ZN_ = REFJOB == 30
# GE_ = REFJOB == 32
# BR2_= REFJOB == 35
# RB_ = REFJOB == 37
# SR_ = REFJOB == 38
# Y_  = REFJOB == 39
# ZR_ = REFJOB == 40
# NB_ = REFJOB == 41
# MO_ = REFJOB == 42
# TC_ = REFJOB == 43
# RU_ = REFJOB == 44
# RH_ = REFJOB == 45
# PD_ = REFJOB == 46
# AG_ = REFJOB == 47
# SN_ = REFJOB == 50
# CS_ = REFJOB == 55
# BA_ = REFJOB == 56
# W_  = REFJOB == 74
# RE_ = REFJOB == 75
# OS_ = REFJOB == 76
# IR_ = REFJOB == 77
# PT_ = REFJOB == 78
# AU_ = REFJOB == 79
#
#
# ###################
# # cell_arg = As(concat('[[',AX,',',AY,',',AZ,'],['
# #                          ,BX,',',BY,',',BZ,'],['
# #                          ,CX,',',CY,',',CZ,']]'),'cell')
# #
# # atoms_args = [As(concat('[',GROUPCONCAT(NUMBER),']'),'numbers')             # numbers
# #               ,As(concat('[[',GROUPCONCAT(X),'],['
# #                           ,GROUPCONCAT(Y),'],['
# #                           ,GROUPCONCAT(Z),']]'),'posT')                     # positions (transpose)
# #               ,cell_arg # cell
# #               ,As(concat('[',GROUPCONCAT(MAGMOM),']'),'magmoms')            # magmoms
# #               ,As(concat('[',GROUPCONCAT(CONSTRAINED),']'),'constrained')]  # constrained
# #
#
# def RAND_(x : float) -> typ.Any:
#     """Random fraction of data, chosen by Bool(id + <number> % <number>)"""
#     x = round(x,2)
#     r = random.randint(0,100000)
#     assert 0 <= x <= 1
#
#     if   x==1:         return '1'
#     elif x==0:         return 'not 1'
#     elif x < 0.5:   LOW = True
#     else:
#      x      = 1 - x
#      LOW = False
#
#     harmonic = map(lambda y: 1./y,range(1,105))
#
#     error = float('inf')
#     for i,h in enumerate(harmonic):
#         diff = x-h
#         if abs(diff)>error:
#             if LOW: return Not(Mod(JOB+r,i))
#             else:   return Mod(JOB+r,i)
#         else: error = abs(diff)
#     raise ValueError('Pick a more reasonable value for RAND than'+str(x))
#
#
# ##Misc Joins#
# atoms_tables = joiner(struct,{atoms:atom__atoms,cell:atoms__cell})



################
#OLD STUFF NEED TO REWRITE UNCOMMENT AND FIX AS NEEDED
#################
# atom2       = Table('atom')
# atoms2      = Table('atoms')
# cell2       = Table('cell')
# calc2       = Table('calc')
# job2        = Table('job')
# element2    = Table('element')
# refstoich2  = Table('refstoich')
# refeng2     = Table('refeng')
#
# final_table_joins = {calc  : calc.id  == job.calc_id
#
#                     finalatoms          = Table('toms : atoms.id == job.finalatoms')
#                     ,cell  : cell.id  == atoms.cell_id}
#
# finalatom_table_joins = {calc  : calc.id  == job.calc_id
#                         ,atoms : atoms.id == job.finalatoms
#                         ,cell  : cell.id  == atoms.cell_id
#                         ,atom  : atom.atoms_id == job.finalatoms}
#
#
# final_table      = joiner(job,final_table_joins)
# final_atom_table = joiner(job,finalatom_table_joins)
#
# #Parent Child Table
# parent_job          = Table('job')
# parent_atoms        = Table('atoms')
# parent_cell         = Table('cell')
# parent_calc         = Table('calc')
#
#
# parent_child_table  = joiner(
#                             joiner(
#                                 joiner(job
#                                 ,{parent_job          : [job.parent      == parent_job.storage_directory, 'LEFT']})
#                                 ,{calc                : [calc.id         == job.calc_id,                  'INNER']
#                                 ,atoms                : [atoms.id        == job.finalatoms,               'INNER']
#                                 ,cell                 : [cell.id         == atoms.cell_id,                'INNER']
#                                 ,parent_calc          : [parent_calc.id  == parent_job.calc_id,            'LEFT']
#                                 ,parent_atoms         : [parent_atoms.id == parent_job.finalatoms,         'LEFT']}
#                                 )
#                             ,{parent_cell          : [parent_cell.id   == parent_atoms.cell_id, 'LEFT']}
#                         )
#
# def get_table_names(joined_tables : sql.Join) -> typ.List[str]:
#     left = joined_tables.left
#     right = joined_tables.right
#     if isinstance(left,sql.Table) and isinstance(right,sql.Table):
#         return [left._name, right._name]
#     elif isinstance(left,sql.Table):
#         return [left._name]+get_table_names(right)
#     else:
#         return get_table_names(left)+[right._name]
#
# td = {'atom':atom,'atoms':atoms,'cell':cell,'calc':calc,'job':job # TABLE
#      ,'refeng':refeng,'composition':composition}                                            # DICTIONARY
#



import pickle as pickle#External Modules
import os,json,glob
import fireworks # type: ignore
################################################################################

#################################
# Auxillary functions for scripts
#################################
def merge_dicts(listDicts):
    import itertools
    return dict(itertools.chain.from_iterable([x.items() for x in listDicts])) #presumes no overlap in keys

def get_cluster():
    import os
    hostname = os.environ['HOSTNAME'].lower()
    if  'sh'in hostname or  'gpu-' in hostname:
        sher = os.environ['SHERLOCK']
        if   sher == '1':  return 'sherlock'
        elif sher == '2': return 'sherlock2'
        else: raise ValueError('What cluster is this? '+sher)

    elif    'su'      in hostname: return 'suncat'
    elif    'nid'     in hostname: return 'nersc'
    else: raise ValueError("clusterRoot could not parse HOSTNAME: %s"%hostname)

def get_number_of_nodes():
    """
    Identify which parallel processor is executing this code
    """
    import os
    pwd = os.getcwd()

    if os.getenv('SHERLOCK')=='2':
        #On Sherlock 2
        with open('world.size','w') as f:
            number_of_nodes = int(os.getenv('SLURM_JOB_NUM_NODES'))
            f.write(str(number_of_nodes))
    elif '/nfs/slac/g/suncatfs' in pwd:
        with open('world.size','w') as f:
            procs = os.getenv('LSB_HOSTS').split()
            number_of_nodes = int(len(procs)/8)
            f.write(str(number_of_nodes))
    else:
        number_of_nodes = 1
    return number_of_nodes

def rank():
    """
    Identify which parallel processor is executing this code
    """
    import sys
    # Check for special MPI-enabled Python interpreters:
    if '_gpaw' in sys.builtin_module_names:
        import _gpaw  # type: ignore      # http://wiki.fysik.dtu.dk/gpaw
        world = _gpaw.Communicator()
    elif '_asap' in sys.builtin_module_names:
        import _asap # type: ignore    # http://wiki.fysik.dtu.dk/asap, can't import asap3.mpi here (import deadlock)
        world = _asap.Communicator()
    elif 'asapparallel3' in sys.modules: # type: ignore    # Older version of Asap
        import asapparallel3 # type: ignore
        world = asapparallel3.Communicator()
    elif 'Scientific_mpi' in sys.modules:
        from Scientific.MPI import world # type: ignore
    else:
        from ase.parallel import DummyMPI # type: ignore
        world = DummyMPI()# This is a standard Python interpreter:
    rank = world.rank #size = world.size

    return rank

#############################################
# Universally-applicable functions for scripts
#############################################

def write_script(listOfFuncs,mainFunc):
    """
    Takes a list of functions and writes them to script.py
    One function is specified as the function called by __main__
    """
    import inspect

    with open('script.py','w') as f:
        for fnc in listOfFuncs:
            f.write('\n\n'+inspect.getsource(fnc))
        f.write("\n\nif __name__ == '__main__': %s()\n\n"%mainFunc.__name__)

    os.system('chmod 755 script.py')


def writeMetaScript(dftcode):
    if dftcode == 'gpaw':
        script = """#!/bin/bash
# Slurm parameters
NTASKS=`echo $SLURM_TASKS_PER_NODE|tr '(' ' '|awk '{print $1}'`
NNODES=`scontrol show hostnames $SLURM_JOB_NODELIST|wc -l`
NCPU=`echo " $NTASKS * $NNODES " | bc`
# load gpaw-specific paths
source /scratch/users/ksb/gpaw/paths.bash
# run parallel gpaw
mpirun -n $NCPU gpaw-python script.py"""

    elif dftcode == 'quantumespresso':
        script = """#!/bin/bash\npython2 script.py"""

    with open('metascript.sh','w') as f: f.write(script)

    os.system('chmod 755 metascript.sh')


def initialize():
    """
    Does three things to initialize (any) job:
        1. Delete old error and output files for restarted jobs
        2. Load job input parameters
        3. Write initial Atoms object to init.traj
    """
    import json,os,glob
    import ase.io as aseio # type: ignore
    import ase.parallel as asepar # type: ignore

    def keepNewest(string):                 # Step 1
        listPth = glob.glob(string)
        ordered = sorted([(os.path.getmtime(pth),pth) for pth in listPth])
        for t,pth in ordered[:-1]:
            os.remove(pth)

    if rank()==0:
        keepNewest('*.error')
        keepNewest('*.out')

    try:   os.remove('result.json')
    except OSError: pass
    try:   os.remove('runtime.json')
    except OSError: pass

    with asepar.paropen('params.json','r') as f: prms  = json.loads(f.read())    # Step 2
    atoms   = make_atoms(prms)
    atoms.set_pbc([1,1,1])

    aseio.write('init.traj',atoms)         # Step 3

    return prms,atoms

###########################
# Atoms object manipulation
###########################

def traj_to_json(atoms):
    """
    Serialize an Atoms object in a human readable way
    """
    import ase, json
    def roundfloat(x):
        output =  round(float(x),3)
        return abs(output) if  output == 0 else output
    atomdata = []
    const    = atoms.constraints
    if const: const_list = const[0].get_indices().tolist()
    else:     const_list = []

    atoms.wrap()
    for a in atoms: atomdata.append({'number'		: int(a.number)
                                    ,'x'			: roundfloat(a.x)
                                    ,'y'			: roundfloat(a.y)
                                    ,'z'			: roundfloat(a.z)
                                    ,'magmom'		: roundfloat(a.magmom)
                                    ,'tag'			: int(a.tag)
                                    ,'constrained'	: int(a.index in const_list)
                                    ,'index'		: int(a.index)})

    out = {'cell': [[roundfloat(x) for x in xx] for xx in atoms.get_cell().tolist()]
          ,'atomdata':atomdata}

    return json.dumps(out)


def json_to_traj(raw_json):
    """
    Inverse of traj_to_json
    """
    from ase.constraints import FixAtoms
    import numpy as np
    import ase, json

    raw_atoms = json.loads(raw_json)
    try:
        atom_data = raw_atoms['atom_data']
    except KeyError:
        atom_data = raw_atoms['atomdata']

    pos = np.array([[a[q] for a in atom_data] for q in ['x','y','z']]).T

    fix = FixAtoms([a['index'] for a in atom_data if a['constrained']])

    atoms = ase.Atoms(numbers     = [a['number'] for a in atom_data]
                     ,cell        = raw_atoms['cell']
                     ,positions   = pos
                     ,magmoms     = [a['magmom'] for a in atom_data]
                     ,tags        = [a['tag'] for a in atom_data]
                     ,constraint  = fix)
    return atoms

def make_atoms(params):
    import pickle as pickle # type: ignore
    try:
        return json_to_traj(params['inittraj'])
    except KeyError:
        return pickle.loads(str(params['inittraj_pckl']))


def cellpar_to_cell(cellpar, ab_normal=(0, 0, 1), a_direction=None):
    """Return a 3x3 cell matrix from cellpar=[a,b,c,alpha,beta,gamma]. Angles must be in degrees.
    The returned cell is orientated such that a and b are normal to `ab_normal` and a is parallel to the projection of `a_direction` in the a-b plane.
    Default `a_direction` is (1,0,0), unless this is parallel to `ab_normal`, in which case default `a_direction` is (0,0,1).
    The returned cell has the vectors va, vb and vc along the rows. The cell will be oriented such that va and vb are normal to `ab_normal` and va will be along the projection of `a_direction` onto the a-b plane.
    """
    import numpy as np # type: ignore
    from numpy import pi, sin, cos, arccos, sqrt, dot
    def unit_vector(x): #Return a unit vector in the same direction as x.
        y = np.array(x, dtype='float')
        return y / np.linalg.norm(y)
    if a_direction is None:
        if np.linalg.norm(np.cross(ab_normal, (1, 0, 0))) < 1e-5: a_direction = (0, 0, 1)
        else: a_direction = (1, 0, 0)
    ad = np.array(a_direction) # Define rotated X,Y,Z-system, with Z along ab_normal and X along
    Z = unit_vector(ab_normal) # the projection of a_direction onto the normal plane of Z.
    X = unit_vector(ad - dot(ad, Z) * Z)
    Y = np.cross(Z, X)
    # Express va, vb and vc in the X,Y,Z-system
    alpha, beta, gamma = 90., 90., 90.
    if isinstance(cellpar, (int, float)): a = b = c = cellpar
    elif len(cellpar) == 1:     a = b = c = cellpar[0]
    elif len(cellpar) == 3: a, b, c = cellpar
    else:  a, b, c, alpha, beta, gamma = cellpar
    # Handle orthorhombic cells separately to avoid rounding errors
    eps = 2 * np.spacing(90.0, dtype=np.float64)  # around 1.4e-14
    if abs(abs(alpha) - 90) < eps: cos_alpha = 0.0 # alpha
    else:  cos_alpha = cos(alpha * pi / 180.0)
    if abs(abs(beta) - 90) < eps: cos_beta = 0.0 # beta
    else: cos_beta = cos(beta * pi / 180.0)
    if abs(gamma - 90) < eps: # gamma
        cos_gamma,sin_gamma = 0.0,1.0
    elif abs(gamma + 90) < eps:
        cos_gamma,sin_gamma = 0.0,-1.0
    else:
        cos_gamma = cos(gamma * pi / 180.0)
        sin_gamma = sin(gamma * pi / 180.0)
    va = a * np.array([1, 0, 0]) # Build the cell vectors
    vb = b * np.array([cos_gamma, sin_gamma, 0])
    cx = cos_beta
    cy = (cos_alpha - cos_beta * cos_gamma) / sin_gamma
    cz = sqrt(1. - cx * cx - cy * cy)
    vc = c * np.array([cx, cy, cz])
    abc = np.vstack((va, vb, vc)) # Convert to the Cartesian x,y,z-system
    T = np.vstack((X, Y, Z))
    return dot(abc, T) #cell

def makeCalc(p):
    """Make a relaxation calculator for GPAW/QE given input parameters"""
    import json

    def makeGPAWcalc():
        from gpaw import (GPAW,PW,Davidson,Mixer,MixerSum,FermiDirac,setup_paths,PoissonSolver) # type: ignore

        if p['psp'] == 'oldpaw':
            setup_paths.insert(0, '/scratch/users/ksb/gpaw/oldpaw/gpaw-setups-0.6.6300/')
            psp = 'paw'
        else: psp = p['psp']

        if p['spinpol'] or p['hund']:
            mixer = MixerSum(beta=p['mixing'],nmaxold=p['nmix'],weight=100)
        else:
            mixer = Mixer(beta=p['mixing'],nmaxold=p['nmix'],weight=100)

        if p['hund']:
            occupations   = FermiDirac(0.01,fixmagmom=True)
            poissonsolver = None # broken??? PoissonSolver(eps=1e-12)
        else:
            occupations   = FermiDirac(p['sigma'])
            poissonsolver = None

        return GPAW(mode         = PW(p['pw'],force_complex_dtype=p['hund'])
                    ,xc          = p['xc']
                    ,kpts        = p['kpts']
                    ,spinpol     = p['spinpol']
                    ,convergence = {'energy':p['econv']} #eV/electron
                    ,mixer       = mixer
                    ,maxiter       = p['maxstep']
                    ,nbands        = p['nbands']
                    ,occupations   = occupations
                    ,setups        = psp
                    ,eigensolver   = 'rmm-diis' if p['hund'] else Davidson(5)
                    ,poissonsolver = poissonsolver
                    ,hund          = p['hund']
                    ,txt           = 'log'
                    ,symmetry      = 'off')

    def makeQEcalc():
        import json
        from espresso import espresso # type: ignore

        pspDict =   {'sherlock' : {'gbrv15pbe':'/home/vossj/suncat/psp/gbrv1.5pbe'}
                    ,'sherlock2': {'gbrv15pbe':'/home/users/vossj/suncat/psp/gbrv1.5pbe'}
                    ,'suncat'   : {'gbrv15pbe':'/nfs/slac/g/suncatfs/sw/external/esp-psp/gbrv1.5pbe'}}
        pspPath =  pspDict[get_cluster()][p['psp']]
        number_of_nodes = get_number_of_nodes()
        return espresso( pw         = p['pw']
                        ,dw         = p['dw']
                        ,xc         = p['xc']
                        ,kpts       = p['kpts']
                        ,spinpol    = p['spinpol']
                        ,convergence=   {'energy':      p['econv']
                                        ,'mixing':      p['mixing']
                                        ,'mixing_mode': p['mixingtype']
                                        ,'nmix':        p['nmix']
                                        ,'maxsteps':    p['maxstep']
                                        ,'diag':        'david'}
                        ,nbands            = p['nbands']
                        ,sigma             = p['sigma']
                        ,dipole            = {'status': p['dipole']}
                        ,outdir            = 'calcdir'
                        ,startingwfc       = 'atomic+random'
                        ,tot_magnetization = 'hund' if p['hund'] else -1
                        ,nspin             = -2 if p['hund'] else None
                        ,psppath           = pspPath
                        ,parflags          = '-npool {}'.format(number_of_nodes)
                        ,mode              = 'scf'
                        ,output     = {'removesave':True})

    if   p['dftcode']=='gpaw': return makeGPAWcalc()
    else:                      return makeQEcalc()

def makeQEVCcalc(p):
    import json
    from espresso import espresso # type: ignore
    pspDict =   {'sherlock': {'gbrv15pbe':'/home/vossj/suncat/psp/gbrv1.5pbe'}
                ,'sherlock2': {'gbrv15pbe':'/home/users/vossj/suncat/psp/gbrv1.5pbe'}
                ,'suncat':   {'gbrv15pbe':'/nfs/slac/g/suncatfs/sw/external/esp-psp/gbrv1.5pbe'}}
    pspPath =  pspDict[get_cluster()][p['psp']]
    number_of_nodes = get_number_of_nodes()
    return  espresso( pw            = p['pw']
                    ,dw             = p['dw']
                    ,xc             = p['xc']
                    ,kpts           = p['kpts']
                    ,nbands         = p['nbands']
                    ,dipole         = {'status': p['dipole']}
                    ,sigma          = p['sigma']
                    ,mode           = 'vc-relax'
                    ,cell_dynamics  = 'bfgs'
                    ,opt_algorithm  = 'bfgs'
                    ,cell_factor    = p['cell_factor']
                    ,cell_dofree    = p['cell_dofree']
                    ,spinpol        = p['spinpol']
                    ,outdir         = 'calcdir'
                    ,output         = {'removesave':True}
                    ,psppath        = pspPath
                    ,parflags       = '-npool {}'.format(number_of_nodes)
                    ,convergence    =   {'energy':  p['econv']
                                        ,'mixing':  p['mixing']
                                        ,'nmix':    p['nmix']
                                        ,'maxsteps':p['maxstep']
                                        ,'diag':    'david'})


def makeQEvibcalc(p):
    import json
    from espresso.vibespresso import vibespresso # type: ignore

    pspDict =   {'sherlock': {'gbrv15pbe':'/home/vossj/suncat/psp/gbrv1.5pbe'}
                ,'sherlock2': {'gbrv15pbe':'/home/users/vossj/suncat/psp/gbrv1.5pbe'}
                ,'suncat':   {'gbrv15pbe':'/nfs/slac/g/suncatfs/sw/external/esp-psp/gbrv1.5pbe'}}
    pspPath =  pspDict[get_cluster()][p['psp']] #this must be done here because we don't know what cluster job will be run on
    number_of_nodes = get_number_of_nodes()
    return vibespresso( pw           = p['pw']
                        ,dw          = p['dw']
                        ,xc          = p['xc']
                        ,kpts        = p['kpts']
                        ,spinpol     = p['spinpol']
                        ,dipole      = {'status': p['dipole']}
                        ,convergence =   {'energy':  p['econv']
                                         ,'mixing':  p['mixing']
                                         ,'nmix':    p['nmix']
                                         ,'maxsteps':p['maxstep']
                                         ,'diag':    'david'}
                        ,nbands      = p['nbands']
                        ,sigma       = p['sigma']
                        ,outdir      = 'calcdir'
                        ,startingwfc = 'atomic+random'
                        ,psppath     = pspPath
                        ,parflags    = '-npool {}'.format(number_of_nodes)
                        ,output      = {'removesave':True,'wf_collect':False})

def makeQEnebCalc(p):
    import json
    from espresso.multiespresso import multiespresso # type: ignore
    pspDict =   {'sherlock': {'gbrv15pbe':'/home/vossj/suncat/psp/gbrv1.5pbe'}
                ,'sherlock2': {'gbrv15pbe':'/home/users/vossj/suncat/psp/gbrv1.5pbe'}
                ,'suncat':   {'gbrv15pbe':'/nfs/slac/g/suncatfs/sw/external/esp-psp/gbrv1.5pbe'}}
    pspPath =  pspDict[get_cluster()][p['psp']] #this must be done here because we don't know what cluster job will be run on

    return multiespresso(ncalc        = p['images']
                        ,outdirprefix ='neb'
                        ,pw           = p['pw']
                        ,dw           = p['dw']
                        ,xc           = p['xc']
                        ,kpts         = p['kpts']
                        ,spinpol      = p['spinpol']
                        ,dipole       = {'status': p['dipole']}
                        ,sigma        = p['sigma']
                        ,nbands       = p['nbands']
                        ,psppath      = pspPath
                		,output       = {'avoidio':False,
                                         'removewf':True,
                                         'wf_collect':False}
                        ,convergence  =  {'energy':  p['econv']
                                         ,'mixing':  p['mixing']
                                         ,'nmix':    p['nmix']
                                         ,'mixing_mode': p['mixingtype']
                                         ,'maxsteps':p['maxstep']
                                         ,'diag':    'david'} )

def cFunc(prms):
    """
    Selects the appropriate calculator-making function given
    properties of the parameter dictionary
    """
    dftcode,jobkind =prms['dftcode'],prms['jobkind']
    relax = prms['jobkind'] in ['latticeopt','relax','bulkmod','fbl','xc']
    if relax :                return makeCalc
    elif jobkind =='vcrelax': return makeQEVCcalc
    elif jobkind == 'vib':    return makeQEvibcalc
    elif jobkind == 'neb':    return makeQEnebCalc
    else: NotImplementedError('jobkind {} has no calculator function'.format(jobkind))


def optimize_pos(atoms,calc,fmax, restart = 'qn.pckl'):
    import ase.optimize as aseopt # type: ignore

    atoms.set_calculator(calc)
    dyn = aseopt.BFGS(atoms=atoms, logfile='qn.log', trajectory='qn.traj',restart = restart)
    dyn.run(fmax=fmax)

def traj_details(atoms):
    """ Returns dictionary summary of an (optimized) Atoms object """
    import pickle as pickle # type: ignore
    import numpy as np # type: ignore

    try: mag = atoms.get_magnetic_moments()
    except: mag = np.array([0]*len(atoms))
    return {'finaltraj':traj_to_json(atoms)
            ,'finalpos_pckl':pickle.dumps(atoms.get_positions())
            ,'finalcell_pckl':pickle.dumps(atoms.get_cell())
            ,'finalmagmom_pckl':pickle.dumps(mag)}

def log(params):
    """
    Execute at the end of all scripts. Copies info to external storage, where it should be added to a database
    """
    import os,shutil,getpass,time,sys,json
    kwargs = json.loads(params.get('kwargs','{}'))

    #####################
    # Auxillary functions
    #--------------------
    def safeMkdir(newDir): # type: ignore
        try:            os.makedirs(newDir)
        except OSError: pass # may fail due to multiple threads or already existing

    def safeCopy(path,name): # type: ignore
        try:            shutil.copyfile(path,name)
        except IOError: pass

    def clusterRoot(): # type: ignore
        hostname = os.environ['HOSTNAME'].lower()
        if      'sh'    in hostname or  'gpu-' in hostname:    return '/scratch/users/ksb/share/jobs'
        elif    'su'    in hostname:                           return '/nfs/slac/g/suncatfs/ksb/share/jobs'
        elif    'nid'   in hostname:                           return '/global/cscratch1/sd/krisb/share/jobs'
        else: raise ValueError


    ######################################################
    print("Collecting environment data...")
    #-----------------------------------------------------
    public_storage    = clusterRoot()+'/'  # type: ignore

    user              = getpass.getuser()
    timestamp         = time.time()
    working_directory = os.getcwd()
    script            = sys.argv[0].split('/')[-1] # e.g. opt.py

    ######################################################
    print("Agglomerating data to be logged in database...")
    #-----------------------------------------------------
    newDir      = public_storage+'%s/%s/'%(user,str(timestamp).replace('.','')) #name of new directory

    allCols     = ['deleted','user',    'timestamp', 'working_directory','storage_directory','kwargs']
    binds       = [0,         user,  int(timestamp),  working_directory,  newDir,          json.dumps(kwargs)]
    runtime     = dict(zip(allCols,binds))

    #################################################
    print("Creating and Populating copy directory...")
    #------------------------------------------------
    safeMkdir(newDir) # type: ignore

    with open(newDir+'runtime.json','w') as f: f.write(json.dumps(runtime))       # Write runtime info to file

    for root, dirs, files in os.walk(working_directory, topdown=False):
        for name in files:
            originalPath = os.path.join(root, name)
            if  name == script:                           # uniform naming system for 'opt.py' script
                safeCopy(originalPath,newDir+'script.py') # type: ignore
            elif '.err' in name:                          # uniform naming system for error file
                safeCopy(originalPath,newDir+'myjob.err') # type: ignore
            elif '.out' in name:                          # uniform naming system from out file
                safeCopy(originalPath,newDir+'myjob.out') # type: ignore
            elif os.stat(originalPath).st_size < 1e9: safeCopy(originalPath,newDir+name)  # type: ignore

    os.system('chmod -R 755 %s'%newDir)

    return 0
############################
############################
# JobKind-specific functions
############################
############################

###########################
# OptimizeLattice functions
###########################

def OptimizeLatticeScript():
    import pickle as pickle
    import json,os,ase,shutil
    import scipy.optimize as opt    # type: ignore
    import ase.parallel   as asepar # type: ignore

    global energies,lparams

    #######################
    print("Initializing...")
    #----------------------
    params,initatoms = initialize()      # Remove old .out/.err files, load from fw_spec, and write 'init.traj'
    shutil.copy('init.traj','out.traj')  # Initial guess for atomic coordinates inside get_bulk_energy reads from 'out.traj'
    energies,lparams = [],[]             # type: ignore

    if rank()==0:
        for d in ['qn.traj','qn.log','lattice_opt.log','energy_forces.pckl']:          # Remove partially completed calculations that may still be held over from failed job
            if os.path.exists(d): os.remove(d)
            print('Removed existing file ',d)

    ################################
    print("Loading initial guess...")
    #-------------------------------

    try:
        with asepar.paropen('lastguess.json','r') as f: iGuess = json.loads(f.read())
        print('\tread lastguess from lastguess.json: ',iGuess)
    except:
        iGuess = get_init_guess(params['structure'],initatoms.get_cell())
        print('\tgenerating initial guess from get_init_guess(): ',iGuess)

    ########################################
    print("Optimizing cell and positions...")
    #---------------------------------------
    optimizedLatticeParams  = opt.fmin(get_bulk_energy,iGuess,args=(params,),ftol=1,xtol=params['xtol'])
    print('Optimized lattice params: ',optimizedLatticeParams)

    ################################
    print("Storing Results...")
    #-------------------------------

    if rank() == 0:
        optAtoms    = ase.io.read('out.traj') #read optimized cell and positions
        with open('energy_forces.pckl') as f: eng,forces=pickle.loads(str(f.read()))
        resultDict  = merge_dicts([params,traj_details(optAtoms)
                                ,{'raw_energy':     eng
                                ,'forces_pckl':     pickle.dumps(forces)
                                ,'latticeopt_pckl': pickle.dumps(zip(energies,lparams))}])

        with open('result.json', 'w') as f: f.write(json.dumps(resultDict))
        log(params)
    return 0

def get_init_guess(structure,cell):
    """
    Convert cell into parameters a,b,c,alpha,beta,gamma
    Depending on structure, return sufficient information required to reconstruct cell (e.g. only 'a' is needed if the structure is known cubic)
    """
    import numpy as np # type: ignore

    def angle(v1,v2): return np.arccos(np.dot(v1,np.transpose(v2))/(np.linalg.norm(v1)*np.linalg.norm(v2)))
    a,b,c            = np.linalg.norm(cell[0]),np.linalg.norm(cell[1]),np.linalg.norm(cell[2])
    alpha,beta,gamma = angle(cell[1],cell[2]), angle(cell[0],cell[2]), angle(cell[0],cell[1])

    if   structure in ['fcc','bcc','fcc-unit','bcc-unit','rocksalt','cubic','diamond','cscl','zincblende','rocksalt-unit','zincblende-unit','M3N-unit','antibixbyite']: return [a]
    elif structure in ['hexagonal','hcp','Li3N']: return [a,c]
    elif structure in ['triclinic']: return [a,b,c,alpha,beta,gamma]
    else: raise ValueError('Bad entry in "structure" field for Atoms object info dictionary: '+structure)

def get_bulk_energy(latticeParams,params):
    #For a given set of bravais lattice parameters, optimize atomic coordinates and return minimum energy
    import pickle as pickle
    import ase,json # type: ignore
    import ase.parallel as asepar # type: ignore
    import ase.io as aseio # type: ignore
    global energies,lparams

    with asepar.paropen('lastguess.json','w') as f: f.write(json.dumps(list(latticeParams)))
    atomsInitUnscaled = make_atoms(params)
    atoms = from_params(atomsInitUnscaled,latticeParams,params['structure'])
    optimize_pos(atoms,makeCalc(params),params['fmax'])
    energy,forces = atoms.get_potential_energy(),atoms.get_forces()
    energies.append(energy);lparams.append(latticeParams)
    with asepar.paropen('lattice_opt.log','a') as logfile: logfile.write('%s\t%s\n' %(energy,latticeParams))
    aseio.write('out.traj',atoms)
    with open('energy_forces.pckl','w') as f: f.write(pickle.dumps((energy,forces)))
    return energy

def from_params(atomsInput,cellParams,structure):
    """
    Params is a list of 1 to 6 numbers (a,b,c,alpha,beta,gamma).
    ANGLES OF INPUT ARE IN RADIANS, ASE CELLS WANT ANGLES IN DEGREES
    Depending on structure, we can construct the cell from a subset of these parameters
    """
    import math     as m

    a = cellParams[0]
    if   structure in ['cubic','cscl','fcc-unit','bcc-unit','zincblende-unit','rocksalt-unit','M3N-unit','antibixbyite']:  cell = [a,a,a,90,90,90]
    elif structure in ['fcc','diamond','zincblende','rocksalt']: cell = [a,a,a,60,60,60]
    elif structure in ['bcc']:                                   cell = [a,a,a,109.47122,109.47122,109.47122]
    elif structure in ['hcp','hexagonal','Li3N']:                       cell = [a,a,cellParams[1],90,90,120]                                                                                          # Input is assumed to be two parameters, a and c
    elif structure in ['triclinic']:                             cell = [cellParams[0],cellParams[1],cellParams[2],m.degrees(cellParams[3]),m.degrees(cellParams[4]),m.degrees(cellParams[5])] # You should really be using VC relax for this....
    else: raise NotImplementedError('from_params(atomsInput,cellParams,structure) cannot handle unknown structure = '+structure)
    atoms = atomsInput.copy()
    atoms.set_cell(cellpar_to_cell(cell),scale_atoms=True)

    return atoms

###########################
# BulkModulus functions
###########################
def BulkModulusScript():
    import json,base64,copy,os
    import ase,matplotlib # type: ignore
    import numpy        as np # type: ignore
    import ase.eos      as aseeos # type: ignore
    import ase.parallel as asepar # type: ignore
    matplotlib.use('Agg')
    #######################
    print("Initializing...")
    #----------------------

    params,optAtoms = initialize()  # Remove old .out/.err files, load from fw_spec, and write 'init.traj'

    optCell,vol,eng = optAtoms.get_cell() , [],[]
    strains         = np.linspace(1 - params['strain'],1 + params['strain'],9)

    if rank()==0:
        for d in ['qn.traj','qn.log']:          # Remove partially completed calculations that may still be held over from failed job
            if os.path.exists(d): os.remove(d)
            print('Removed existing file ',d)

    ##################################################
    print("Calculating energies at various strains...")
    #-------------------------------------------------

    for i, strain in enumerate(strains):
        atoms = optAtoms.copy()
        atoms.set_cell(optCell*strain,scale_atoms=True)
        optimize_pos(atoms,makeCalc(params),params['fmax'])

        volume,energy = atoms.get_volume(),atoms.get_potential_energy()
        vol.append(copy.deepcopy(volume));eng.append(copy.deepcopy(energy))

    ################################
    print("Analyzing dE/dV curve...")
    #-------------------------------

    aHat,quadR2 = quad_fit(np.array(copy.deepcopy(vol)),np.array(copy.deepcopy(eng)))

    try:
        eos = aseeos.EquationOfState(vol,eng)
        v0, e0, b = eos.fit()
        eos.plot(filename='bulk-eos.png',show=False)
        b0= b/ase.units.kJ*1e24                                     #GPa: use this value if EOS doesn't fail
        with open('bulk-eos.png', 'rb') as f: img = base64.b64encode(f.read())

    except ValueError:                                  # too bad of a fit for ASE to handle
        b0 = aHat*2*vol[4]*160.2                        # units: eV/A^6 * A^3 * 1, where 1 === 160.2 GPa*A^3/eV
        img = None

    ################################
    print("Storing Results...")
    #-------------------------------

    resultDict  = merge_dicts([params,traj_details(optAtoms)
                                ,{'bulkmod':b0,'bfit':quadR2,'bulkmodimg_base64':img,'voleng_json':json.dumps(zip(vol,eng))}])

    with open('result.json', 'w') as outfile:   outfile.write(json.dumps(resultDict))

    if rank()==0: log(params)
    return 0

def quad_fit(xIn,yIn):
    """ Input x vector: units A^3, Input y vector: units eV """
    import numpy            as np # type: ignore
    import scipy.optimize   as opt # type: ignore

    # Center data around 4th data point
    x = xIn-xIn[4]; y = yIn-yIn[4]
    def model(a):  return a*np.square(x)
    def errVec(a): return a*np.square(x) - y   #create fitting function of form mx+b
    aHat, success = opt.leastsq(errVec, [0.1])
    yhat    = model(aHat)
    ybar    = np.sum(y)/len(y)          # or sum(y)/len(y)
    ssresid = np.sum((yhat-y)**2)
    sstotal = np.sum((y - ybar)**2)    # or sum([ (yi - ybar)**2 for yi in y])
    r2      = 1 - (ssresid / float(sstotal)) #
    return float(aHat[0]),float(r2)

###########################
# XC Contribs functions
##########################
def XCcontribsScript():
    import json,ase,copy # type: ignore
    from gpaw        import restart # type: ignore
    from gpaw.xc.bee import BEEFEnsemble # type: ignore
    params,atoms = initialize()  # Remove old .out/.err files, load from fw_spec, and write 'init.traj'
    pbeParams = copy.deepcopy(params)
    pbeParams['xc']='PBE'
    atoms.set_calculator(makeCalc(pbeParams))
    atoms.get_potential_energy()
    atoms.calc.write('inp.gpw', mode='all')

    atoms,calc = restart('inp.gpw', setups='sg15', xc='mBEEF'
                            , convergence={'energy': 5e-4}, txt='mbeef.txt')
    atoms.get_potential_energy()
    beef        = BEEFEnsemble(calc)
    xcContribs  = beef.mbeef_exchange_energy_contribs()
    ################################
    print("Storing Results...")
    #-------------------------------
    ase.io.write('final.traj',atoms)
    resultDict  = merge_dicts([params,traj_details(ase.io.read('final.traj'))
                                ,{'xc_contribs': xcContribs.tolist()}])
    with open('result.json', 'w') as outfile:   outfile.write(json.dumps(resultDict))
    if rank()==0: log(params)
    return 0

#################
# Relax functions
#################

def RelaxScript():
    import pickle as pickle
    import ase,json,os # type: ignore
    import ase.parallel   as asepar # type: ignore

    #######################
    print("Initializing...")
    #----------------------

    params,initatoms = initialize()  # Remove old .out/.err files, load from fw_spec, and write 'init.traj'

    if rank()==0:
        if os.path.exists('qn.traj'):
            if os.stat('qn.traj').st_size > 100:
                initatoms = ase.io.read('qn.traj')
                print('\treading from qn.traj...')
            else:
                os.remove('qn.traj')
                print('removed empty qn.traj')
        if os.path.exists('qn.log') and os.stat('qn.log').st_size < 100:
            os.remove('qn.log')
            print('\tremoved empty qn.log...')

    #######################
    print("Optimizing positions...")
    #----------------------
    optimize_pos(initatoms,makeCalc(params),params['fmax'])

    ############################
    print("Storing Results...")
    #--------------------------
    ase.io.write('final.traj',initatoms)
    e0 = initatoms.get_potential_energy()
    f0 = initatoms.get_forces()
    optAtoms = ase.io.read('final.traj')
    resultDict  = merge_dicts([params,traj_details(optAtoms)
                                ,{'raw_energy': e0
                                ,'forces_pckl':pickle.dumps(f0)} ])

    with open('result.json', 'w') as f: f.write(json.dumps(resultDict))
    if rank()==0: log(params)
    return 0


#####################
# Vibration Functions
#####################
def VibScript():
    from ase.vibrations       import Vibrations # type: ignore
    import pickle as pickle
    import glob,json,ase,os # type: ignore

    #######################
    print("Initializing...")
    #----------------------

    params,atoms = initialize()  # Remove old .out/.err files, load from fw_spec, and write 'init.traj'

    prev = glob.glob('*.pckl') #delete incomplete pckls - facilitating restarted jobs

    if rank()==0:
        for p in prev:
            if os.stat(p).st_size < 100: os.remove(p)

    atoms.set_calculator(makeQEvibcalc(params))
    vib = Vibrations(atoms,delta=params['delta'],indices=json.loads(params['vibids']))

    vib.run(); vib.write_jmol()

    ##########################
    print("Storing Results...")
    #-------------------------

    vib.summary(log='vibrations.txt')

    with open('vibrations.txt','r') as f: vibsummary = f.read()

    ase.io.write('final.traj',atoms)
    optAtoms = ase.io.read('final.traj')

    vib_energies,vib_frequencies = vib.get_energies(),vib.get_frequencies()

    resultDict  = merge_dicts([params,  traj_details(optAtoms),{'vibfreqs_pckl': pickle.dumps(vib_frequencies)
                                        ,'vibsummary':vibsummary
                                        ,'vibengs_pckl':pickle.dumps(vib_energies)}])

    with open('result.json', 'w') as outfile:   outfile.write(json.dumps(resultDict))
    with open('result.json', 'r') as outfile: json.loads(outfile.read()) #test that dictionary isn't 'corrupted'
    if rank()==0: log(params)
    return 0


#####################
# VCRelax Functions
#####################

def VCRelaxScript():
    import ase,json,pickle,os # type: ignore
    #######################
    print("Initializing...")
    #----------------------
    params,atoms = initialize()  # Remove old .out/.err files, load from fw_spec, and write 'init.traj'

    if not os.path.exists('intermediate.traj'):
        ##########################################
        print("Running VC Relax for first time...")
        #-----------------------------------------
        atoms.set_calculator(makeQEVCcalc(params))
        energy = atoms.get_potential_energy()   # Trigger espresso to be launched
        ase.io.write('intermediate.traj',atoms.calc.get_final_structure())

    ###########################################
    print("Running VC Relax for second time...")
    #------------------------------------------
    atoms = ase.io.read('intermediate.traj')
    atoms.set_calculator(makeQEVCcalc(params))
    energy = atoms.get_potential_energy() #trigger espresso to be launched
    ase.io.write('final.traj',atoms.calc.get_final_structure())

    ################################
    print("Storing Results...")
    #-------------------------------
    e0,f0   = atoms.get_potential_energy(),atoms.get_forces()
    atoms   = ase.io.read('final.traj')

    resultDict  = merge_dicts([params,traj_details(atoms),
                                {'raw_energy':  e0
                                ,'forces_pckl': pickle.dumps(f0)}])

    with open('result.json', 'w') as outfile:   outfile.write(json.dumps(resultDict))
    with open('result.json', 'r') as outfile: json.loads(outfile.read()) #test that dictionary isn't 'corrupted'
    if rank()==0: log(params)
    return 0

#####################
# FBL Script
#####################

def FBLScript():
    import ase,json,base64,copy,os,matplotlib, fnmatch # type: ignore
    import numpy        as np # type: ignore
    from ase.constraints import FixBondLength # type: ignore
    import pickle
    matplotlib.use('Qt5Agg')
    #######################
    print("Initializing...")
    #----------------------

    params,optAtoms = initialize()  # Remove old .out/.err files, load from fw_spec, and write 'init.traj'

    #Check for work that has already been done will use the last completed step as starting point
    step = 0
    if rank() == 0:
        step_files = fnmatch.filter(os.walk('.').next()[2], 'step_*_.traj')
        if not step_files == []:
            step_list = sorted(map(lambda x: int(x.split('_')[1]),step_files), reverse = True)
            for step in step_list:
                try:
                    optAtoms = ase.io.read('step_{0}_.traj'.format(step))
                    step += 1
                    print("Restarting at step {0}".format(step))
                    break
                except:
                    pass

    #Initialize the energy_cut_off of steps and step size
    energy_cut_off  = params['energy_cut_off']
    step_size       = params['step_size']
    fix             = params['fix'] if fix in params.keys() else 0.5

    #Initialize bond distance and fix the bond length
    bonded_inds = json.loads(params['bonded_inds'])
    optAtoms.set_constraint(optAtoms.constraints+[FixBondLength(*bonded_inds)])


    d_init, energy_array, atoms_array = optAtoms.get_distance(*bonded_inds), [], []
    if step == 0:
        print("Starting with inittraj_pckl, bond distance = {0}".format(d_init))
        d_array = [d_init]
    else:
        d_init += -step_size
        print("Beginning with bond distance = {0}".format(d_init))
        complete_steps = np.arange(0,step)
        atoms_array += map(lambda x: ase.io.read('step_{0}_.traj'.format(x)),complete_steps)
        energy_array += map(lambda x: x.get_potential_energy(),atoms_array)
        d_array = map(lambda x: x.get_distance(*bonded_inds),atoms_array)

    #Create the rest of the bond distance array that will be iterated through
    d_array += sorted(np.arange(0.3,(np.floor((d_init)/step_size)*step_size+0.01),step_size),reverse=True)

    if rank()==0:
        for d in ['qn.traj','qn.log']:          # Remove partially completed calculations that may still be held over from failed job
            if os.path.exists(d): os.remove(d)
            print('Removed existing file ',d)

    ##################################################
    print("Reducing the bond distance and relaxing atoms")
    #-------------------------------------------------
    if round(d_init,1) == 0.2:
        print('Bond distance already at or below 0.3 Angstrom, no additional steps will be taken')
    else:
        atoms = optAtoms.copy()
        for d_curr in d_array[step:]:
            step_curr = d_array.index(d_curr)
            if step_curr>0 and energy_array[-1]-energy_array[0]>energy_cut_off:
                d_array = d_array[:step_curr]
                print('Energy cutoff hit at step {0}'.format(step_curr))
                break
            atoms.set_distance(*bonded_inds,distance = d_curr, fix = 0.5)
            optimize_pos(atoms,makeCalc(params),params['fmax'], restart = None)
            ase.io.write('step_{0}_.traj'.format(step_curr),atoms)
            atoms_array.append(atoms.copy())
            energy_array.append(atoms.get_potential_energy())
            print('Step {0} completed: Energy = {1}; Bond Distance = {2}'.format(step_curr, energy_array[-1], atoms.get_distance(*bonded_inds)))

    ################################
    print("Analyzing energy vs bond length curve...")
    #-------------------------------
    d_array.reverse(); energy_array.reverse(); atoms_array.reverse()
    TS_step, TS_dis, TS_energy, TS_atoms, TS_barrier, rxn_energy = get_FBL_TS_energy(d_array, energy_array, atoms_array)
    img = FBL_plotter(d_array, energy_array, TS_step = TS_step)

    ################################
    print("Storing Results...")
    #-------------------------------
    ase.io.write('TS.traj',TS_atoms)
    fbl_dict = {'fbl_barrier':TS_barrier,'rxn_energy':rxn_energy
                ,'TS_bond_length':TS_dis,'TS_atoms': pickle.dumps(TS_atoms)
                ,'TS_energy':TS_energy,'fblimg_base64':img}
    resultDict  = merge_dicts([params,traj_details(TS_atoms), fbl_dict,])

    with open('result.json', 'w') as outfile:   outfile.write(json.dumps(resultDict))

    if rank()==0: log(params)
    return 0

def get_FBL_TS_energy(d_list, e_list, atoms_list, min_d = 0.7):
    from scipy.interpolate import CubicSpline # type: ignore
    import numpy as np # type: ignore
    from copy import deepcopy

    d_list =deepcopy(d_list); e_list = deepcopy(e_list); atoms_list = deepcopy(atoms_list)

    #Fit data to a cubic spline for more precise identification of peak
    cs = CubicSpline(d_list,e_list)
    xs = np.arange(min(d_list),max(d_list),0.01)
    diff_e = np.diff(cs(xs))

    #Find peaks from instanteous slopes in the cubic spline fit
    peaks = []
    for i, slope in enumerate(diff_e>0):
        if i< len(diff_e)-1:
            if slope and not diff_e[i+1]>0:
                peaks.append(i+1)

    #Identify the TS as the step with the closest energy to the peak energy
    if peaks == []:
        max_peak_ind = np.argmax(cs(xs))
    else:
        max_peak_ind = peaks[np.argmax([cs(xs)[i] for i in peaks])]

    TS_step = np.argmin(abs(xs[max_peak_ind]-d_list))

    #Store the data for the transition state
    TS_atoms = atoms_list[TS_step]
    TS_dis = d_list[TS_step]
    TS_energy = e_list[TS_step]
    TS_barrier = TS_energy - e_list[np.argmax(d_list)]

    #Find bonded state through similar algorithm
    troughs = []
    for i, slope in enumerate(diff_e<0):
        if i< len(diff_e)-1:
            if slope and not diff_e[i+1]<0:
                troughs.append(i+1)
    if troughs ==[]:
        troughs = [0]
        print('No bonded state found!!!')
    bonded_dis = np.min([xs[i] for i in troughs])
    bonded_step = np.argmin(abs(np.array(d_list)-bonded_dis))
    bonded_energy = e_list[bonded_step]
    rxn_energy = bonded_energy - e_list[np.argmin(d_list)]

    if TS_dis < min_d:
        print('Warning: very small ts bond length')
    return TS_step, TS_dis, TS_energy, TS_atoms, TS_barrier, rxn_energy

def FBL_plotter(d_list, e_list, TS_step = None):
    import matplotlib
    matplotlib.use('TKAgg')
    import matplotlib.pyplot as plt # type: ignore
    import numpy as np # type: ignore
    import base64
    fig  = plt.figure()
    e_list = np.array(e_list) - e_list[-1]
    plt.plot(d_list,e_list)
    if TS_step == None:
        plt.plot(d_list[TS_step],e_list[TS_step],marker ='o',color='r')
    plt.xlabel('Bond length (Angstroms)')
    plt.ylabel('Energy (eV)')
    plt.title('Barrier = %1.4f eV'%(e_list[TS_step]))
    plt.savefig('./fbl_plot.png')
    with open('./fbl_plot.png', 'rb') as f: img = base64.b64encode(f.read())
    return img

#####################
# IN PROGRESS
#####################


def NebScript():

    import os

    params,initatoms = initialize()  # Remove old .out/.err files, load from fw_spec, and write 'init.traj'
    n = params['images']
    if not os.path.exists('neb1.traj'): neb = initializeNeb(params)
    else: neb = reloadNeb()

    while True:
        if check_unconverged(params):
            address_unconverged(params)
        else:
            statuses = [check_images_near_peaks(i) for i in range(n)]
            if statuses.count('peak')==1: return finalize(statuses.index('peak')) #converged
            elif statuses.count('near_peak')==2: split_neb(*[s for s in statuses if s=='nearpeak'])
            else: run_neb()

def initializeNeb(params):
    """
    read in information from job param dictionary
    """

    import pickle as pickle # type: ignore
    from ase.neb import NEB # type: ignore

    #######################
    print("Initializing...")
    #----------------------

    finalatoms = pickle.loads(str(params['finaltraj_pckl']))
    write('initial.traj',initatoms)
    write('final.traj',finalatoms)
    images = [initatoms]
    for i in range(n):
        images.append(initatoms.copy())
    images.append(finalatoms)
    neb = NEB(images,k=params['k'])
    neb.interpolate()
    return neb

def run_neb(params):
    """
    Run NEB
    """
    from ase.optimize import FIRE # type: ignore
    from ase.io import Trajectory # type: ignore

    #######################
    print("Running...")
    #----------------------

    m = makeQEnebCalc(params)
    m.set_neb(neb)
    qn = FIRE(neb,logfile='qn.log')
    for j in range(1,n+1):
        traj = Trajectory('neb%d.traj' % j, 'w', images[j])
        qn.attach(traj)

    qn.run(fmax=params['fmax'])

def reloadNeb(params):
    """
    Use existing files in working directory to restart NEB
    """

    from ase.io import write,read # type: ignore

    ################################
    print("Reloading NEB images...")
    #-------------------------------

    images = [read('initial.traj')]
    for i in range(1,n+1):
        images.append(read('neb%d.traj'%i))
    images.append(read('final.traj'))
    neb = NEB(images,k=params['k'])
    return neb

def check_unconverged():
    """
    Check if NEB is unconverged
    """

    #######################
    print("Checking convergence")
    #----------------------

    raise NotImplementedError

def address_unconverged():
    """
    Deal with unconverged NEB
    """

    #########################################
    print("Addressing unconverged NEB...")
    #----------------------------------------

    raise NotImplementedError

def check_image_near_peak(img):
    """

    """

    #########################################
    print("Check if image %d is at a peak..."%img)
    #----------------------------------------

    raise NotImplementedError

def finalize(peak_img):
    """
    Final touches after NEB has converged
    """

    ################################
    print("Storing Results...")
    #-------------------------------

    raise NotImplementedError

def split_neb(peak1,peak2):
    """
    Address a NEB that seems to have two transition states
    """

    #########################################
    print("Addressing NEB with two peaks...")
    #----------------------------------------

    raise NotImplementedError

#####################
#####################
#####################

def DosScript():
    raise NotImplementedError

#############
# FIRETASKS
#############

def executeAndCollect():
    os.system('./metascript.sh')
    try:
        with open('result.json','r') as f:
            resultDictjson = f.read()
    except FileNotFoundError:
        pass

    return fireworks.core.firework.FWAction(stored_data={'resultjson':resultDictjson})

@fireworks.utilities.fw_utilities.explicit_serialize
class OptimizeLattice(fireworks.core.firework.FiretaskBase):
    def run_task(self,fw_spec):

        params = fw_spec['params']

        with open('params.json','w') as f: f.write(json.dumps(params))

        write_script([initialize  # Universal
                    ,traj_to_json
                    ,json_to_traj
                    ,make_atoms
                    ,cFunc(params)
                    ,optimize_pos
                    ,traj_details
                    ,log
                    ,OptimizeLatticeScript # Jobkind-specific
                    ,get_init_guess
                    ,get_bulk_energy
                    ,from_params
                    ,cellpar_to_cell
                    ,rank
                    ,merge_dicts # Auxillary
                    ,get_cluster
                    ,get_number_of_nodes
                    ],OptimizeLatticeScript)

        writeMetaScript(params['dftcode'])

        return executeAndCollect()



@fireworks.utilities.fw_utilities.explicit_serialize
class GetBulkModulus(fireworks.core.firework.FiretaskBase):
    def run_task(self,fw_spec):


        params = fw_spec['params']

        with open('params.json','w') as f: f.write(json.dumps(params))

        write_script([initialize  # Universal
                    ,traj_to_json
                    ,json_to_traj
                    ,make_atoms
                    ,cFunc(params)
                    ,optimize_pos
                    ,traj_details
                    ,log
                    ,BulkModulusScript # Jobkind-specific
                    ,quad_fit
                    ,rank
                    ,merge_dicts # Auxillary
                    ,get_cluster
                    ,get_number_of_nodes
                    ],BulkModulusScript)

        writeMetaScript(params['dftcode'])

        os.system('./metascript.sh')

        with open('result.json','r') as f: resultDictjson = f.read()

        return fireworks.core.firework.FWAction(stored_data={'resultjson':resultDictjson})

@fireworks.utilities.fw_utilities.explicit_serialize
class GetXCcontribs(fireworks.core.firework.FiretaskBase):
    def run_task(self,fw_spec):

        params = fw_spec['params']

        with open('params.json','w') as f: f.write(json.dumps(params))

        write_script([initialize         # Universal
                    ,traj_to_json
                    ,json_to_traj
                    ,make_atoms
                    ,cFunc(params)
                    ,traj_details
                    ,log
                    ,XCcontribsScript   # Jobkind-specific
                    ,merge_dicts         # Auxillary
                    ,rank
                    ,get_cluster
                    ,get_number_of_nodes
                    ],XCcontribsScript)

        writeMetaScript(params['dftcode'])

        return executeAndCollect()

#################################################################
@fireworks.utilities.fw_utilities.explicit_serialize
class Relax(fireworks.core.firework.FiretaskBase):
    def run_task(self,fw_spec):

        params = fw_spec['params']

        with open('params.json','w') as f: f.write(json.dumps(params))

        write_script([initialize    # Universal
                    ,traj_to_json
                    ,json_to_traj
                    ,make_atoms
                    ,cFunc(params)
                    ,optimize_pos
                    ,log
                    ,traj_details
                    ,RelaxScript    # Jobkind-specific
                    ,rank           # Auxillary
                    ,merge_dicts
                    ,get_cluster
                    ,get_number_of_nodes
                    ],RelaxScript)

        writeMetaScript(params['dftcode'])

        return executeAndCollect()

@fireworks.utilities.fw_utilities.explicit_serialize
class Vibrations(fireworks.core.firework.FiretaskBase):
    def run_task(self,fw_spec):

        params = fw_spec['params']

        with open('params.json','w') as f: f.write(json.dumps(params))

        write_script([initialize         # Universal
                    ,traj_to_json
                    ,json_to_traj
                    ,make_atoms
                    ,cFunc(params)
                    ,optimize_pos
                    ,traj_details
                    ,log
                    ,VibScript  # Jobkind-specific
                    ,rank
                    ,merge_dicts         # Auxillary
                    ,get_cluster
                    ,get_number_of_nodes
                    ],VibScript)

        writeMetaScript(params['dftcode'])

        return executeAndCollect()

@fireworks.utilities.fw_utilities.explicit_serialize
class FBL(fireworks.core.firework.FiretaskBase):
    def run_task(self,fw_spec):

        params = fw_spec['params']

        with open('params.json','w') as f: f.write(json.dumps(params))

        write_script([initialize         # Universal
                    ,traj_to_json
                    ,json_to_traj
                    ,make_atoms
                    ,cFunc(params)
                    ,optimize_pos
                    ,traj_details
                    ,log
                    ,FBLScript          # Jobkind-specific
                    ,get_FBL_TS_energy
                    ,FBL_plotter
                    ,rank
                    ,merge_dicts         # Auxillary
                    ,get_cluster
                    ,get_number_of_nodes
                    ],FBLScript)

        writeMetaScript(params['dftcode'])

        return executeAndCollect()

@fireworks.utilities.fw_utilities.explicit_serialize
class DOS(fireworks.core.firework.FiretaskBase):
    def run_task(self,fw_spec):

        params = fw_spec['params']

        with open('params.json','w') as f: f.write(json.dumps(params))

        write_script([initialize         # Universal
                    ,traj_to_json
                    ,json_to_traj
                    ,make_atoms
                    ,cFunc(params)
                    ,optimize_pos
                    ,traj_details
                    ,log
                    ,DosScript          # Jobkind-specific
                    ,rank
                    ,merge_dicts         # Auxillary
                    ,get_cluster
                    ,get_number_of_nodes
                    ],DosScript)

        writeMetaScript(params['dftcode'])

        return executeAndCollect()

@fireworks.utilities.fw_utilities.explicit_serialize
class NEB(fireworks.core.firework.FiretaskBase):
    def run_task(self,fw_spec):

        params = fw_spec['params']

        with open('params.json','w') as f: f.write(json.dumps(params))

        write_script([initialize         # Universal
                    ,traj_to_json
                    ,json_to_traj
                    ,make_atoms
                    ,cFunc(params)
                    ,traj_details
                    ,log
                    ,NebScript          # Jobkind-specific
                    ,rank
                    ,merge_dicts         # Auxillary
                    ,get_cluster
                    ,get_number_of_nodes
                    ],NebScript)

        writeMetaScript(params['dftcode'])

        return executeAndCollect()

@fireworks.utilities.fw_utilities.explicit_serialize
class VCRelax(fireworks.core.firework.FiretaskBase):
    def run_task(self,fw_spec):

        params = fw_spec['params']

        with open('params.json','w') as f: f.write(json.dumps(params))

        write_script([initialize         # Universal
                    ,traj_to_json
                    ,json_to_traj
                    ,make_atoms
                    ,cFunc(params)
                    ,optimize_pos
                    ,traj_details
                    ,log
                    ,VCRelaxScript      # Jobkind-specific
                    ,rank
                    ,merge_dicts         # Auxillary
                    ,get_cluster
                    ,get_number_of_nodes
                    ],VCRelaxScript)

        writeMetaScript(params['dftcode'])

        return executeAndCollect()

def vasp_script():
    import ase.calculators.vasp as vasp_calculator # type: ignore
    import json,multiprocessing,time,os,subprocess

    params,atoms = initialize()      # Remove old .out/.err files, load from fw_spec, and write 'init.traj'
    #############################################
    print("Initializing VASP calculator...")
    #-----------------------------------------

    calc  =  vasp_calculator.Vasp(encut   = params['pw']
                                ,setups   = json.loads(params['psp']) # dictionary {'Cu':'_s'}
                                ,xc       = params['xc']
                                ,gga      = params['gga']      # 'BF'
                                ,luse_vdw = params['luse_vdw'] # True
                                ,zab_vdw  = params.get('zab_vdw')
                                ,kpts     = json.loads(params['kpts'])
                                ,npar     = 1                  # use this if you run on one node (most calculations).  see suncat confluence page for optimal setting
                                ,kpar     = 1
                                ,gamma    = params['gamma']    # True,  # Gamma-centered (defaults to Monkhorst-Pack)
                                ,ismear   = 0                  # assume gaussian smearing, we can make this parameter if necessary
                                ,algo     = params['algo']     #'fast',
                                ,nelm     = params['maxstep']
                                ,sigma    = params['sigma']
                                ,ibrion   = params['ibrion']   #2
                                ,nelmdl   = params['nelmdl']
                                ,isif     = 3 if params['jobkind']=='vcrelax' else 2
                                ,ediffg   = -params['fmax']                        # forces
                                ,ediff    = params['econv']                        # energy conv. both of these are for the internal relaxation, ie nsw
                                ,prec     = params['prec']                         #'Accurate'
                                ,nsw      = params['ionic_steps']                  # use ASE ?
                                ,ispin    = 2 if params['spinpol'] else 1
                                ,lreal    = params['lreal']                        # automatically decide to do real vs recip space calc
                                ,ldipol   = params['dipole']                       #True,
                                ,lvhar    = True                                   #True,
                                ,dipol    = params.get('dipol')                    #(0.5,0.5,0.5),
                                ,idipol   = 3 if params['dipole'] else None
                                ,icharg   = 1)                                     # start from CHGCAR if icharg = 1

    atoms.set_calculator(calc)

    ############################################
    print("Generating VASP Input files with ASE")
    #------------------------------------------
    try:
        proc = multiprocessing.Process(target=atoms.get_potential_energy)
        proc.start();time.sleep(10)
        if proc.is_alive(): proc.terminate()
    except IOError: pass #on nersc, we get io error

    #######################
    print("Running VASP natively...")
    #----------------------

    loader = {'sherlock':  "ml vasp/5.4.1.intel.gpu"
             ,'sherlock2': "ml chemistry vasp/5.4.1"
             ,'nersc':      "module load vasp-tpc/5.4.1"}

    vaspout, vasperr = subprocess.Popen(loader[get_cluster()]+';which vasp_std;srun -n 32 vasp_std', stdout=subprocess.PIPE, stderr=subprocess.PIPE,shell=True).communicate()
    print('VASP output:\n',vaspout)
    with open('OUTCAR','r') as f:
        if 'General timing and accounting informations for this job' in f.read():
            if rank()==0: log(params)
            return 0
    raise ValueError(vasperr)

@fireworks.utilities.fw_utilities.explicit_serialize
class VaspJob(fireworks.core.firework.FiretaskBase):
    def run_task(self,fw_spec):
        import os
        params = fw_spec['params']
        with open('params.json','w') as f: f.write(json.dumps(params))
        write_script([initialize,rank,get_cluster,make_atoms,log,vasp_script],vasp_script)
        os.system('python script.py')
        return 0


########################################################################
# WHAT WE EXPORT
########################################################################
script_dict = {'latticeopt': OptimizeLattice()
             ,'bulkmod': GetBulkModulus(),'vasp':VaspJob()
             ,'relax': Relax(),'vcrelax': VCRelax()
             ,'fbl': FBL(),'dos': DOS(),'neb': NEB()
             ,'vib': Vibrations(),'xc': GetXCcontribs()}



#External Modules
import typing as typ
if typ.TYPE_CHECKING:
    from catalog.jobs.repeat       import RepeatChecker
    import catalog.jobs.repeat,catalog.jobs.scripts
import time,random,json,os,math,pickle
import datetime as d    # type: ignore
import ase              # type: ignore
import numpy as np      # type: ignore
import fireworks as fw  # type:ignore
# Internal Modules
import catalog.jobs.cluster as clust
from catalog.jobs              import scripts
from catalog.misc.sql_utils    import *
from catalog.misc.print_parse  import np_to_str
from catalog.misc.atoms        import json_to_traj
# User-required files
import catalog.catalog_config as config # type: ignore

"""
Notes: ?
"""

class Job(object):
    """
    Class for all DFT calculations.
    """


    def __init__(self
                ,params : typ.Dict[str,typ.Any]
                ) -> None:

        assert all([isinstance(k,str) for k in params])

        ##############################################
        # store some params as fields for a 'shortcut'
        #---------------------------------------------
        self.jobkind    = params['jobkind']
        self.dftcode    = params['dftcode']

        #For old jobs, to_string is a useful method for the database
        #These jobs have pickletrajs still so the below try/except will catch both
        #In future old jobs will need their pickle_trajs converted to jsons
        try:
            self.atoms      = json_to_traj(params['inittraj'])
        except KeyError:
            print('FOUND JOB WITH PICKLE TRAJ')
            self.atoms      = pickle.loads(params['inittraj_pckl'].encode('latin1'), encoding='latin1')
        self.finalatoms = params.get('finaltraj')

        self.params = params



    def __str__(self) -> str: return self.to_string('exact')

    def _str_calc(self) -> str:
        calc_keys = ['dftcode','xc','pw','kpts','fmax','psp','econv','xtol'
                    ,'strain','dw','sigma','nbands','mixing','nmix','gga'
                    ,'luse_vdw','zab_vdw','nelmdl','gamma','dipol','algo'
                    ,'ibrion','prec','ionic_steps','lreal','lvhar','diag'
                    ,'spinpol','dipole','maxstep','delta','mixingtype'
                    ,'bonded_inds','energy_cut_off','step_size','springs'
                    ,'hund']

        return '_'.join(['%s=%s'%(k,str(self.params.get(k))) for k in calc_keys])

    def to_string(self
                 ,granularity : str = 'exact'
                 ) -> str:
        """
        Represent the job with a string that can be checked for equality
        """
        s = self._str_calc()
        s += '__'+self._traj_to_str(granularity)
        # for nebs, add traj_to_str of finalatoms?
        return s

    def _traj_to_str(self
                    ,granularity : str
                    ) -> str:
        """Represent an Atoms object with a string: two levels of granularity"""

        # Convert constraints to a string
        #--------------------------------
        cnst = str(self.atoms.constraints)

        if granularity == 'exact': #FULL INFORMATION
            p = np_to_str(self.atoms.get_positions())
            c = np_to_str(self.atoms.get_cell())
            n = json.dumps(self.atoms.get_atomic_numbers().tolist())
            m = np_to_str(self.atoms.get_initial_magnetic_moments(),matrix=False)

            return '_'.join([n,p,c,cnst,m])

        elif granularity == 'graph': #LOSE PRECISE POSITION + CELL INFO, LOSE MAGMOM INFO
            # raise NotImplementedError
            return ''
            #from catalog.structure.graphatoms import GraphMaker
            #return GraphMaker(jmol=True).make_unweighted_canonical(atoms=self.atoms)

        else:
            raise NotImplementedError('Granularity not supported '+granularity)
            return ''

    def submit(self
              ,repeat_checker : "RepeatChecker"
              ,walltime       : typ.Optional[float]          = None
              ,clust_str      : typ.Optional[str]            = None
              ,nodes          : typ.Optional[int]            = None
              ,cores          : typ.Optional[int]            = None
              ) -> int:
        """
        Submit the job described by params to Fireworks
        """


        if clust_str is None:  cluster  = clust.cluster_dict[config.allocate(self,1)[0]]             # For now we assume every workflow has only one firework
        else:                  cluster  = clust.str2Cluster(self,clust_str,1)[0] # we pass in a string for Cluster
        if walltime is None:   walltime = config.guessTime(self)
        if nodes    is None:   nodes    = config.guessNodes(self)



        if not repeat_checker.check_repeat(self):
            repeat_checker.add_incomplete_job(self) # Modify RepeatChecker object
            launchpad     = fw.LaunchPad.from_file(config.LAUNCHPAD_YAML)
            wflow         = self._wflow(walltime, cluster, nodes, cores)
            time.sleep(2) # folder names = timestamp to nearest second
            return launchpad.add_wf(wflow)
        else:
            return 0

    def _standard_script(self) -> typ.Any:
        """
        Helpful docstring
        """
        if self.dftcode == 'vasp': key = 'vasp'
        else:                      key = self.jobkind
        return scripts.script_dict[key]

    def _wflow(self
              ,walltime  : int
              ,cluster   : clust.Cluster
              ,nodes     : int
              ,cores     : int
              ) -> "fw.WorkFlow":
        """
        Create a fireworks workflow
        """

        #Check if any non-default values for job submission are selected
        #If one is not specified the catalog_config functions will be used

        timestamp = '_'+d.datetime.now().strftime('%Y_%m_%d_%H_%M_%S')

        print("Submited to {0} for {1} Hours".format(cluster.name, walltime))

        fw1 = fw.Firework([self._standard_script()]
                        ,name = self.jobkind
                        ,spec = {'params':          self.params
                                ,'_fworker':        cluster.fworker.name
                                ,'_queueadapter':   cluster.qfunc(walltime,nodes,cores)
                                ,'_launch_dir':     os.path.join(cluster.get_launchdir(),self.jobkind+timestamp)})

        return fw.Workflow([fw1],name=self.jobkind)




from ase.atoms import Atoms, string2symbols   # type: ignore
from ase.units import kB, kJ, kcal, mol   # type: ignore
import numpy as np   # type: ignore

# Sol57LC
sol57_lp    = ['Li_bcc','Na_bcc','K_bcc','Rb_bcc','Ca_fcc','Sr_fcc','Ba_bcc',
               'V_bcc','Nb_bcc','Ta_bcc','Mo_bcc','W_bcc','Fe_bcc',
               'Rh_fcc','Ir_fcc','Ni_fcc','Pd_fcc','Pt_fcc','Cu_fcc',
               'Ag_fcc','Au_fcc','Al_fcc',
               'C_diamond','Si_diamond','Ge_diamond','Sn_diamond',
               'LiH_b1','LiF_b1','LiCl_b1','NaF_b1','NaCl_b1','MgO_b1','MgS_b1',
               'CaO_b1','TiC_b1','TiN_b1','ZrC_b1','ZrN_b1','VC_b1','VN_b1',
               'NbC_b1','NbN_b1','FeAl_b2','CoAl_b2','NiAl_b2','BN_b3',
               'BP_b3','BAs_b3','AlN_b3','AlP_b3','AlAs_b3','GaN_b3',
               'GaP_b3','GaAs_b3','InP_b3','InAs_b3','SiC_b3']

assert len(sol57_lp) == 57
#SMS 03/2016
#Sol27LC
sol27_lp = sol57_lp[0:26]
sol27_lp.append('Pb_fcc')
assert len(sol27_lp) == 27
data_info = {
            'sol27_lp':sol27_lp,
            'slab':[],
            'molecule':[],
            'ads':[],
            'dbh24':[],
            'bulk':[]
            }
#SMS 03/2016

# Sol58LC = Sol57LC + Pb_fcc
sol58_lp    = ['Li_bcc','Na_bcc','K_bcc','Rb_bcc','Ca_fcc','Sr_fcc','Ba_bcc',
               'V_bcc','Nb_bcc','Ta_bcc','Mo_bcc','W_bcc','Fe_bcc',
               'Rh_fcc','Ir_fcc','Ni_fcc','Pd_fcc','Pt_fcc','Cu_fcc',
               'Ag_fcc','Au_fcc','Al_fcc','Pb_fcc',
               'C_diamond','Si_diamond','Ge_diamond','Sn_diamond',
               'LiH_b1','LiF_b1','LiCl_b1','NaF_b1','NaCl_b1','MgO_b1','MgS_b1',
               'CaO_b1','TiC_b1','TiN_b1','ZrC_b1','ZrN_b1','VC_b1','VN_b1',
               'NbC_b1','NbN_b1','FeAl_b2','CoAl_b2','NiAl_b2','BN_b3',
               'BP_b3','BAs_b3','AlN_b3','AlP_b3','AlAs_b3','GaN_b3',
               'GaP_b3','GaAs_b3','InP_b3','InAs_b3','SiC_b3']
assert len(sol58_lp) == 58
# Sol53Ec
sol53_coh    = ['Li_bcc','Na_bcc','K_bcc','Rb_bcc','Ca_fcc','Sr_fcc','Ba_bcc',
                'V_bcc','Nb_bcc','Ta_bcc','Mo_bcc','W_bcc','Fe_bcc',
                'Rh_fcc','Ir_fcc','Ni_fcc','Pd_fcc','Pt_fcc','Cu_fcc',
                'Ag_fcc','Au_fcc','Al_fcc',
                'C_diamond','Si_diamond','Ge_diamond','Sn_diamond',
                'LiH_b1','LiF_b1','LiCl_b1','NaF_b1','NaCl_b1','MgO_b1','MgS_b1',
                'CaO_b1','TiC_b1','TiN_b1','ZrC_b1','ZrN_b1','VC_b1','VN_b1',
                'NbC_b1','NbN_b1','BN_b3','BP_b3','AlN_b3','AlP_b3','AlAs_b3',
                'GaN_b3','GaP_b3','GaAs_b3','InP_b3','InAs_b3','SiC_b3']
assert len(sol53_coh) == 53
# Sol54Ec
sol54_coh    = ['Li_bcc','Na_bcc','K_bcc','Rb_bcc','Ca_fcc','Sr_fcc','Ba_bcc',
                'V_bcc','Nb_bcc','Ta_bcc','Mo_bcc','W_bcc','Fe_bcc',
                'Rh_fcc','Ir_fcc','Ni_fcc','Pd_fcc','Pt_fcc','Cu_fcc',
                'Ag_fcc','Au_fcc','Al_fcc','Pb_fcc',
                'C_diamond','Si_diamond','Ge_diamond','Sn_diamond',
                'LiH_b1','LiF_b1','LiCl_b1','NaF_b1','NaCl_b1','MgO_b1','MgS_b1',
                'CaO_b1','TiC_b1','TiN_b1','ZrC_b1','ZrN_b1','VC_b1','VN_b1',
                'NbC_b1','NbN_b1','BN_b3','BP_b3','AlN_b3','AlP_b3','AlAs_b3',
                'GaN_b3','GaP_b3','GaAs_b3','InP_b3','InAs_b3','SiC_b3']
assert len(sol54_coh) == 54
# BM32
bm32        = ['Li_bcc','Na_bcc','Ca_fcc','Sr_fcc','Ba_bcc',
               'Rh_fcc','Pd_fcc','Cu_fcc','Ag_fcc','Al_fcc',
               'C_diamond','Si_diamond','Ge_diamond','Sn_diamond',
               'LiH_b1','LiF_b1','LiCl_b1','NaF_b1','NaCl_b1','MgO_b1',
               'BN_b3','BP_b3','BAs_b3','AlN_b3','AlP_b3','AlAs_b3',
               'GaN_b3','GaP_b3','GaAs_b3','InP_b3','InAs_b3','SiC_b3']
assert len(bm32) == 32
# hcp
hcp_solids_10 = ['Cd_hcp', 'Co_hcp', 'Os_hcp', 'Ru_hcp', 'Zn_hcp',
                 'Zr_hcp', 'Sc_hcp', 'Be_hcp', 'Mg_hcp', 'Ti_hcp']
assert len(hcp_solids_10) == 10
# SE30
se21 = ['Li110','Na110','K110','Rb110','Ba110','Ca111','Sr111',
        'Nb110','Ta110','Mo110','W110','Fe110','Al111','Ni111',
        'Cu111','Rh111','Pd111','Ag111','Ir111','Pt111','Au111']
se9  = ['Mg0001','Zn0001','Cd0001','Sc0001','Ti0001',
        'Co0001','Zr0001','Ru0001','Os0001']
se30 = se21+se9
assert len(se30) == 30
